# Copyright (C) 2003-2005 Alexey "Snake" Nezhdanov
#     modified by Dimitur Kirov <dkirov@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Main xmpp decision making logic. Provides library with methods to assign
different handlers to different XMPP stanzas and namespaces
"""

import inspect
import logging
import sys
import uuid
from xml.parsers.expat import ExpatError

from nbxmpp.modules.activity import Activity
from nbxmpp.modules.adhoc import AdHoc
from nbxmpp.modules.annotations import Annotations
from nbxmpp.modules.attention import Attention
from nbxmpp.modules.blocking import Blocking
from nbxmpp.modules.captcha import Captcha
from nbxmpp.modules.chat_markers import ChatMarkers
from nbxmpp.modules.chatstates import Chatstates
from nbxmpp.modules.correction import Correction
from nbxmpp.modules.delay import Delay
from nbxmpp.modules.discovery import Discovery
from nbxmpp.modules.eme import EME
from nbxmpp.modules.entity_caps import EntityCaps
from nbxmpp.modules.http_auth import HTTPAuth
from nbxmpp.modules.http_upload import HTTPUpload
from nbxmpp.modules.ibb import IBB
from nbxmpp.modules.idle import Idle
from nbxmpp.modules.iq import BaseIq
from nbxmpp.modules.location import Location
from nbxmpp.modules.message import BaseMessage
from nbxmpp.modules.misc import unwrap_carbon
from nbxmpp.modules.misc import unwrap_mam
from nbxmpp.modules.mood import Mood
from nbxmpp.modules.muc import MUC
from nbxmpp.modules.muclumbus import Muclumbus
from nbxmpp.modules.nickname import Nickname
from nbxmpp.modules.omemo import OMEMO
from nbxmpp.modules.oob import OOB
from nbxmpp.modules.openpgp import OpenPGP
from nbxmpp.modules.pgplegacy import PGPLegacy
from nbxmpp.modules.presence import BasePresence
from nbxmpp.modules.pubsub import PubSub
from nbxmpp.modules.receipts import Receipts
from nbxmpp.modules.register import Register
from nbxmpp.modules.security_labels import SecurityLabels
from nbxmpp.modules.software_version import SoftwareVersion
from nbxmpp.modules.tune import Tune
from nbxmpp.modules.user_avatar import UserAvatar
from nbxmpp.modules.vcard_avatar import VCardAvatar
from nbxmpp.namespaces import Namespace
from nbxmpp.plugin import PlugIn
from nbxmpp.protocol import ERR_FEATURE_NOT_IMPLEMENTED
from nbxmpp.protocol import Error
from nbxmpp.protocol import InvalidFrom
from nbxmpp.protocol import InvalidJid
from nbxmpp.protocol import InvalidStanza
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Message
from nbxmpp.protocol import Node
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Presence
from nbxmpp.protocol import Protocol
from nbxmpp.simplexml import NodeBuilder
from nbxmpp.util import get_properties_struct
from nbxmpp.util import INVALID_XML_RX

log = logging.getLogger("nbxmpp.dispatcher")

#: default timeout to wait for response for our id
DEFAULT_TIMEOUT_SECONDS = 25

XML_DECLARATION = "<?xml version='1.0'?>"


# FIXME: ugly
class Dispatcher:
    """
    Why is this here - I needed to redefine Dispatcher for BOSH and easiest way
    was to inherit original Dispatcher (now renamed to XMPPDispatcher). Trouble
    is that reference used to access dispatcher instance is in Client attribute
    named by __class__.__name__ of the dispatcher instance .. long story short:

    I wrote following to avoid changing each client.Dispatcher.whatever() in
    xmpp

    If having two kinds of dispatcher will go well, I will rewrite the
    dispatcher references in other scripts
    """

    def PlugIn(self, client_obj, after_SASL=False, old_features=None):
        if client_obj.protocol_type == "XMPP":
            XMPPDispatcher().PlugIn(client_obj)
        elif client_obj.protocol_type == "BOSH":
            BOSHDispatcher().PlugIn(client_obj, after_SASL, old_features)
        else:
            assert False  # should never be reached

    @classmethod
    def get_instance(cls, *args, **kwargs):
        """
        Factory Method for object creation

        Use this instead of directly initializing the class in order to make
        unit testing much easier.
        """
        return cls(*args, **kwargs)


class XMPPDispatcher(PlugIn):
    """
    Handles XMPP stream and is the first who takes control over a fresh stanza

    Is plugged into NonBlockingClient but can be replugged to restart handled
    stream headers (used by SASL f.e.).
    """

    def __init__(self):
        PlugIn.__init__(self)
        self.handlers = {}
        self._modules = {}
        self._expected = {}
        self._defaultHandler = None
        self._pendingExceptions = []
        self._eventHandler = None
        self._cycleHandlers = []
        self._exported_methods = [
            self.RegisterHandler,
            self.RegisterDefaultHandler,
            self.RegisterEventHandler,
            self.UnregisterCycleHandler,
            self.RegisterCycleHandler,
            self.RegisterHandlerOnce,
            self.UnregisterHandler,
            self.RegisterProtocol,
            self.SendAndCallForResponse,
            self.getAnID,
            self.Event,
            self.send,
            self.get_module,
        ]

    def getAnID(self):
        return str(uuid.uuid4())

    def dumpHandlers(self):
        """
        Return set of user-registered callbacks in it's internal format. Used
        within the library to carry user handlers set over Dispatcher replugins
        """
        return self.handlers

    def restoreHandlers(self, handlers):
        """
        Restore user-registered callbacks structure from dump previously
        obtained via dumpHandlers. Used within the library to carry user
        handlers set over Dispatcher replugins.
        """
        self.handlers = handlers

    def get_module(self, name):
        return self._modules[name]

    def _register_modules(self):
        self._modules["BasePresence"] = BasePresence(self._owner)
        self._modules["BaseMessage"] = BaseMessage(self._owner)
        self._modules["BaseIq"] = BaseIq(self._owner)
        self._modules["EME"] = EME(self._owner)
        self._modules["HTTPAuth"] = HTTPAuth(self._owner)
        self._modules["Nickname"] = Nickname(self._owner)
        self._modules["MUC"] = MUC(self._owner)
        self._modules["Delay"] = Delay(self._owner)
        self._modules["Captcha"] = Captcha(self._owner)
        self._modules["Idle"] = Idle(self._owner)
        self._modules["PGPLegacy"] = PGPLegacy(self._owner)
        self._modules["VCardAvatar"] = VCardAvatar(self._owner)
        self._modules["EntityCaps"] = EntityCaps(self._owner)
        self._modules["Blocking"] = Blocking(self._owner)
        self._modules["PubSub"] = PubSub(self._owner)
        self._modules["Mood"] = Mood(self._owner)
        self._modules["Activity"] = Activity(self._owner)
        self._modules["Tune"] = Tune(self._owner)
        self._modules["Location"] = Location(self._owner)
        self._modules["UserAvatar"] = UserAvatar(self._owner)
        self._modules["OpenPGP"] = OpenPGP(self._owner)
        self._modules["OMEMO"] = OMEMO(self._owner)
        self._modules["Annotations"] = Annotations(self._owner)
        self._modules["Muclumbus"] = Muclumbus(self._owner)
        self._modules["SoftwareVersion"] = SoftwareVersion(self._owner)
        self._modules["AdHoc"] = AdHoc(self._owner)
        self._modules["IBB"] = IBB(self._owner)
        self._modules["Discovery"] = Discovery(self._owner)
        self._modules["ChatMarkers"] = ChatMarkers(self._owner)
        self._modules["Receipts"] = Receipts(self._owner)
        self._modules["OOB"] = OOB(self._owner)
        self._modules["Correction"] = Correction(self._owner)
        self._modules["Attention"] = Attention(self._owner)
        self._modules["SecurityLabels"] = SecurityLabels(self._owner)
        self._modules["Chatstates"] = Chatstates(self._owner)
        self._modules["Register"] = Register(self._owner)
        self._modules["HTTPUpload"] = HTTPUpload(self._owner)

        for instance in self._modules.values():
            for handler in instance.handlers:
                self.RegisterHandler(*handler)

    def _init(self):
        """
        Register default namespaces/protocols/handlers. Used internally
        """
        # FIXME: inject dependencies, do not rely that they are defined by our
        # owner
        self.RegisterNamespace("unknown")
        self.RegisterNamespace(Namespace.STREAMS)
        self.RegisterNamespace(self._owner.defaultNamespace)
        self.RegisterProtocol("iq", Iq)
        self.RegisterProtocol("presence", Presence)
        self.RegisterProtocol("message", Message)
        self.RegisterDefaultHandler(self.returnStanzaHandler)
        self.RegisterEventHandler(self._owner._caller._event_dispatcher)
        self._register_modules()

    def plugin(self, _owner):
        """
        Plug the Dispatcher instance into Client class instance and send initial
        stream header. Used internally
        """
        self._init()
        self._owner.lastErrNode = None
        self._owner.lastErr = None
        self._owner.lastErrCode = None
        if hasattr(self._owner, "StreamInit"):
            self._owner.StreamInit()
        else:
            self.StreamInit()

    def plugout(self):
        """
        Prepare instance to be destructed
        """
        self._modules = {}
        self.Stream.dispatch = None
        self.Stream.features = None
        self.Stream.destroy()
        self._owner = None
        self.Stream = None

    def StreamInit(self):
        """
        Send an initial stream header
        """
        self._owner.Connection.sendqueue = []
        self.Stream = NodeBuilder()
        self.Stream.dispatch = self.dispatch
        self.Stream._dispatch_depth = 2
        self.Stream.stream_header_received = self._check_stream_start
        self.Stream.features = None
        self._metastream = Node("stream:stream")
        self._metastream.setNamespace(self._owner.Namespace)
        self._metastream.setAttr("version", "1.0")
        self._metastream.setAttr("xmlns:stream", Namespace.STREAMS)
        self._metastream.setAttr("to", self._owner.Server)
        self._metastream.setAttr("xml:lang", self._owner.lang)
        self._owner.send("%s%s>" % (XML_DECLARATION, str(self._metastream)[:-2]))

    def _check_stream_start(self, ns, tag, _attrs):
        if ns != Namespace.STREAMS or tag != "stream":
            raise ValueError(
                "Incorrect stream start: " "(%s,%s). Terminating." % (tag, ns)
            )

    def replace_non_character(self, data):
        return INVALID_XML_RX.sub("\ufffd", data)

    def ProcessNonBlocking(self, data):
        """
        Check incoming stream for data waiting

        :param data: data received from transports/IO sockets
        :return:
                1) length of processed data if some data were processed;
                2) '0' string if no data were processed but link is alive;
                3) 0 (zero) if underlying connection is closed.
        """
        # FIXME:
        # When an error occurs we disconnect the transport directly. Client's
        # disconnect method will never be called.
        # Is this intended?
        # also look at transports start_disconnect()
        data = self.replace_non_character(data)
        for handler in self._cycleHandlers:
            handler(self)
        if len(self._pendingExceptions) > 0:
            _pendingException = self._pendingExceptions.pop()
            sys.excepthook(*_pendingException)
            return None
        try:
            self.Stream.Parse(data)
            # end stream:stream tag received
            if self.Stream and self.Stream.has_received_endtag():
                self._owner.disconnect(self.Stream.stream_error)
                return 0
        except ExpatError as error:
            log.error("Invalid XML received from server. Forcing disconnect.")
            log.error(error)
            self._owner.Connection.disconnect()
            return 0
        except ValueError as error:
            log.debug("ValueError: %s", error)
            self._owner.Connection.pollend()
            return 0
        if len(self._pendingExceptions) > 0:
            _pendingException = self._pendingExceptions.pop()
            sys.excepthook(*_pendingException)
            return None
        if len(data) == 0:
            return "0"
        return len(data)

    def RegisterNamespace(self, xmlns):
        """
        Create internal structures for newly registered namespace

        You can register handlers for this namespace afterwards. By default
        one namespace is already registered
        (jabber:client or jabber:component:accept depending on context.
        """
        log.debug('Registering namespace "%s"', xmlns)
        self.handlers[xmlns] = {}
        self.RegisterProtocol("unknown", Protocol, xmlns=xmlns)
        self.RegisterProtocol("default", Protocol, xmlns=xmlns)

    def RegisterProtocol(self, tag_name, proto, xmlns=None):
        """
        Used to declare some top-level stanza name to dispatcher

        Needed to start registering handlers for such stanzas. Iq, message and
        presence protocols are registered by default.
        """
        if not xmlns:
            xmlns = self._owner.defaultNamespace
        log.debug('Registering protocol "%s" as %s(%s)', tag_name, proto, xmlns)
        self.handlers[xmlns][tag_name] = {"type": proto, "default": []}

    def RegisterNamespaceHandler(self, xmlns, handler, typ="", ns="", system=0):
        """
        Register handler for processing all stanzas for specified namespace
        """
        self.RegisterHandler("default", handler, typ, ns, xmlns, system)

    def RegisterHandler(
        self, name, handler, typ="", ns="", xmlns=None, system=False, priority=50
    ):
        """
        Register user callback as stanzas handler of declared type

        Callback arguments:
        dispatcher instance (for replying), incoming return of previous
        handlers. The callback must raise xmpp.NodeProcessed just before return
        if it wants to prevent other callbacks to be called with the same stanza
        as argument _and_, more     importantly     library from returning
        stanza to sender with error set.

        :param name: name of stanza. F.e. "iq".
        :param handler: user callback.
        :param typ: value of stanza's "type" attribute. If not specified any
                value will match
        :param ns: namespace of child that stanza must contain.
        :param xmlns: xml namespace
        :param system: call handler even if NodeProcessed Exception were raised
                       already.
        :param priority: The priority of the handler, higher get called later
        """
        if not xmlns:
            xmlns = self._owner.defaultNamespace

        if not typ and not ns:
            typ = "default"

        log.debug(
            'Registering handler %s for "%s" type->%s ns->%s(%s) priority->%s',
            handler,
            name,
            typ,
            ns,
            xmlns,
            priority,
        )

        if xmlns not in self.handlers:
            self.RegisterNamespace(xmlns)
        if name not in self.handlers[xmlns]:
            self.RegisterProtocol(name, Protocol, xmlns)

        specific = typ + ns
        if specific not in self.handlers[xmlns][name]:
            self.handlers[xmlns][name][specific] = []

        self.handlers[xmlns][name][specific].append(
            {
                "func": handler,
                "system": system,
                "priority": priority,
                "specific": specific,
            }
        )

    def RegisterHandlerOnce(self, name, handler, typ="", ns="", xmlns=None, system=0):
        """
        Unregister handler after first call (not implemented yet)
        """
        # FIXME Drop or implement
        if not xmlns:
            xmlns = self._owner.defaultNamespace
        self.RegisterHandler(name, handler, typ, ns, xmlns, system)

    def UnregisterHandler(self, name, handler, typ="", ns="", xmlns=None):
        """
        Unregister handler. "typ" and "ns" must be specified exactly the same as
        with registering.
        """
        if not xmlns:
            xmlns = self._owner.defaultNamespace
        if not typ and not ns:
            typ = "default"
        if xmlns not in self.handlers:
            return
        if name not in self.handlers[xmlns]:
            return

        specific = typ + ns
        if specific not in self.handlers[xmlns][name]:
            return
        for handler_dict in self.handlers[xmlns][name][specific]:
            if handler_dict["func"] == handler:
                try:
                    self.handlers[xmlns][name][specific].remove(handler_dict)
                    log.debug(
                        'Unregister handler %s for "%s" type->%s ns->%s(%s)',
                        handler,
                        name,
                        typ,
                        ns,
                        xmlns,
                    )
                except ValueError:
                    log.warning(
                        'Unregister failed: %s for "%s" type->%s ns->%s(%s)',
                        handler,
                        name,
                        typ,
                        ns,
                        xmlns,
                    )

    def RegisterDefaultHandler(self, handler):
        """
        Specify the handler that will be used if no NodeProcessed exception were
        raised. This is returnStanzaHandler by default.
        """
        self._defaultHandler = handler

    def RegisterEventHandler(self, handler):
        """
        Register handler that will process events. F.e. "FILERECEIVED" event.
        See common/connection: _event_dispatcher()
        """
        self._eventHandler = handler

    def returnStanzaHandler(self, conn, stanza):
        """
        Return stanza back to the sender with <feature-not-implemented/> error
        set
        """
        if stanza.getType() in ("get", "set"):
            conn._owner.send(Error(stanza, ERR_FEATURE_NOT_IMPLEMENTED))

    def RegisterCycleHandler(self, handler):
        """
        Register handler that will be called on every Dispatcher.Process() call
        """
        if handler not in self._cycleHandlers:
            self._cycleHandlers.append(handler)

    def UnregisterCycleHandler(self, handler):
        """
        Unregister handler that will be called on every Dispatcher.Process()
        call
        """
        if handler in self._cycleHandlers:
            self._cycleHandlers.remove(handler)

    def Event(self, realm, event, data=None):
        """
        Raise some event

        :param realm: scope of event. Usually a namespace.
        :param event: the event itself. F.e. "SUCCESSFUL SEND".
        :param data: data that comes along with event. Depends on event.
        """
        if self._eventHandler:
            self._eventHandler(realm, event, data)
        else:
            log.warning("Received unhandled event: %s", event)

    def dispatch(self, stanza):
        """
        Main procedure that performs XMPP stanza recognition and calling
        apppropriate handlers for it. Called by simplexml
        """

        self.Event("", "STANZA RECEIVED", stanza)

        self.Stream._mini_dom = None

        # Count stanza
        self._owner.Smacks.count_incoming(stanza.getName())

        name = stanza.getName()
        if name == "features":
            self._owner.got_features = True
            self.Stream.features = stanza
        elif name == "error":
            if stanza.getTag("see-other-host"):
                self._owner.got_see_other_host = stanza

        xmlns = stanza.getNamespace()

        if xmlns not in self.handlers:
            log.warning("Unknown namespace: %s", xmlns)
            xmlns = "unknown"
        # features stanza has been handled before
        if name not in self.handlers[xmlns]:
            if name not in ("features", "stream"):
                log.warning("Unknown stanza: %s", stanza)
            else:
                log.debug("Got %s / %s stanza", xmlns, name)
            name = "unknown"
        else:
            log.debug("Got %s / %s stanza", xmlns, name)

        # Convert simplexml to Protocol object
        try:
            stanza = self.handlers[xmlns][name]["type"](node=stanza)
        except InvalidJid:
            log.warning("Invalid JID, ignoring stanza")
            log.warning(stanza)
            return

        own_jid = self._owner.get_bound_jid()
        properties = get_properties_struct(name, own_jid)

        if name == "iq":
            if stanza.getFrom() is None and own_jid is not None:
                stanza.setFrom(own_jid.bare)

        if name == "message":
            # https://tools.ietf.org/html/rfc6120#section-8.1.1.1
            # If the stanza does not include a 'to' address then the client MUST
            # treat it as if the 'to' address were included with a value of the
            # client's full JID.

            to = stanza.getTo()
            if to is None:
                stanza.setTo(own_jid)

            elif not to.bare_match(own_jid):
                log.warning("Message addressed to someone else: %s", stanza)
                return

            if stanza.getFrom() is None:
                stanza.setFrom(own_jid.bare)

            # Unwrap carbon
            try:
                stanza, properties.carbon = unwrap_carbon(stanza, own_jid)
            except (InvalidFrom, InvalidJid) as exc:
                log.warning(exc)
                log.warning(stanza)
                return
            except NodeProcessed as exc:
                log.info(exc)
                return

            # Unwrap mam
            try:
                stanza, properties.mam = unwrap_mam(stanza, own_jid)
            except (InvalidStanza, InvalidJid) as exc:
                log.warning(exc)
                log.warning(stanza)
                return

        typ = stanza.getType()
        if name == "message" and not typ:
            typ = "normal"
        elif not typ:
            typ = ""

        stanza.props = stanza.getProperties()
        log.debug("type: %s, properties: %s", typ, stanza.props)

        _id = stanza.getID()
        processed = False
        if _id in self._expected:
            cb, args = self._expected[_id]
            log.debug("Expected stanza arrived. Callback %s(%s) found", cb, args)
            try:
                if args is None:
                    cb(self, stanza)
                else:
                    cb(self, stanza, **args)
            except NodeProcessed:
                pass
            return

        # Gather specifics depending on stanza properties
        specifics = ["default"]
        if typ and typ in self.handlers[xmlns][name]:
            specifics.append(typ)
        for prop in stanza.props:
            if prop in self.handlers[xmlns][name]:
                specifics.append(prop)
            if typ and typ + prop in self.handlers[xmlns][name]:
                specifics.append(typ + prop)

        # Create the handler chain
        chain = []
        chain += self.handlers[xmlns]["default"]["default"]
        for specific in specifics:
            chain += self.handlers[xmlns][name][specific]

        # Sort chain with priority
        chain.sort(key=lambda x: x["priority"])

        for handler in chain:
            if not processed or handler["system"]:
                try:
                    log.info("Call handler: %s", handler["func"].__qualname__)
                    # Backwards compatibility until all handlers support
                    # properties
                    signature = inspect.signature(handler["func"])
                    if len(signature.parameters) > 2:
                        handler["func"](self, stanza, properties)
                    else:
                        handler["func"](self, stanza)
                except NodeProcessed:
                    processed = True
                except Exception:
                    self._pendingExceptions.insert(0, sys.exc_info())
                    return

        # Stanza was not processed call default handler
        if not processed and self._defaultHandler:
            self._defaultHandler(self, stanza)

    def SendAndCallForResponse(self, stanza, func=None, args=None):
        """
        Put stanza on the wire and call back when recipient replies. Additional
        callback arguments can be specified in args
        """
        _waitid = self.send(stanza)
        self._expected[_waitid] = (func, args)
        return _waitid

    def send(self, stanza, now=False):
        """
        Wrap transports send method when plugged into NonBlockingClient. Makes
        sure stanzas get ID and from tag.
        """
        ID = None
        if isinstance(stanza, Protocol):
            ID = stanza.getID()
            if ID is None:
                stanza.setID(self.getAnID())
                ID = stanza.getID()
            if self._owner._registered_name and not stanza.getAttr("from"):
                stanza.setAttr("from", self._owner._registered_name)

        self._owner.Connection.send(stanza, now)

        # If no ID then it is a whitespace
        if hasattr(self._owner, "Smacks") and ID:
            self._owner.Smacks.save_in_queue(stanza)

        return ID


class BOSHDispatcher(XMPPDispatcher):

    def PlugIn(self, owner, after_SASL=False, old_features=None):
        self.old_features = old_features
        self.after_SASL = after_SASL
        XMPPDispatcher.PlugIn(self, owner)

    def StreamInit(self):
        """
        Send an initial stream header
        """
        self.Stream = NodeBuilder()
        self.Stream.dispatch = self.dispatch
        self.Stream._dispatch_depth = 2
        self.Stream.stream_header_received = self._check_stream_start
        self.Stream.features = self.old_features

        self._metastream = Node("stream:stream")
        self._metastream.setNamespace(self._owner.Namespace)
        self._metastream.setAttr("version", "1.0")
        self._metastream.setAttr("xmlns:stream", Namespace.STREAMS)
        self._metastream.setAttr("to", self._owner.Server)
        self._metastream.setAttr("xml:lang", self._owner.lang)

        self.restart = True
        self._owner.Connection.send_init(after_SASL=self.after_SASL)

    def StreamTerminate(self):
        """
        Send a stream terminator
        """
        self._owner.Connection.send_terminator()

    def ProcessNonBlocking(self, data=None):
        if self.restart:
            fromstream = self._metastream
            fromstream.setAttr("from", fromstream.getAttr("to"))
            fromstream.delAttr("to")
            data = "%s%s>%s" % (XML_DECLARATION, str(fromstream)[:-2], data)
            self.restart = False
        return XMPPDispatcher.ProcessNonBlocking(self, data)

    def dispatch(self, stanza):
        if stanza.getName() == "body" and stanza.getNamespace() == Namespace.HTTP_BIND:

            stanza_attrs = stanza.getAttrs()
            if "authid" in stanza_attrs:
                # should be only in init response
                # auth module expects id of stream in document attributes
                self.Stream._document_attrs["id"] = stanza_attrs["authid"]
            self._owner.Connection.handle_body_attrs(stanza_attrs)

            children = stanza.getChildren()
            if children:
                for child in children:
                    # if child doesn't have any ns specified, simplexml
                    # (or expat) thinks it's of parent's (BOSH body) namespace,
                    # so we have to rewrite it to jabber:client
                    if child.getNamespace() == Namespace.HTTP_BIND:
                        child.setNamespace(self._owner.defaultNamespace)
                    XMPPDispatcher.dispatch(self, child)
        else:
            XMPPDispatcher.dispatch(self, stanza)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Node
from nbxmpp.structs import RSMData


def parse_rsm(stanza: Node) -> RSMData | None:
    stanza = stanza.getTag("set", namespace=Namespace.RSM)
    if stanza is None:
        return None

    after = stanza.getTagData("after") or None
    before = stanza.getTagData("before") or None
    last = stanza.getTagData("last") or None

    first_index = None
    first = stanza.getTagData("first") or None
    if first is not None:
        try:
            first_index = int(first.getAttr("index"))
        except Exception:
            pass

    try:
        count = int(stanza.getTagData("count"))
    except Exception:
        count = None

    try:
        max_ = int(stanza.getTagData("max"))
    except Exception:
        max_ = None

    try:
        index = int(stanza.getTagData("index"))
    except Exception:
        index = None

    return RSMData(
        after=after,
        before=before,
        last=last,
        first=first,
        first_index=first_index,
        count=count,
        max=max_,
        index=index,
    )



from __future__ import annotations

from typing import Any

import collections.abc
import unittest

from nbxmpp.language import basic_filter_languages
from nbxmpp.language import LanguageMap
from nbxmpp.language import LanguageRange
from nbxmpp.language import LanguageTag
from nbxmpp.language import lookup_language


class TestLanguageTag(unittest.TestCase):
    def test_init(self):
        with self.assertRaises(TypeError):
            LanguageTag("foo")

        with self.assertRaises(ValueError):
            LanguageTag(tag="")

        with self.assertRaises(TypeError):
            LanguageTag()

    def test_fromstr_match_str(self):
        tag = LanguageTag.fromstr("de-Latn-DE-1999")
        self.assertEqual("de-latn-de-1999", tag.match_str)

    def test_fromstr_print_str(self):
        tag = LanguageTag.fromstr("de-Latn-DE-1999")
        self.assertEqual("de-Latn-DE-1999", tag.print_str)

    def test___str__(self):
        tag = LanguageTag.fromstr("zh-Hans")
        self.assertEqual("zh-Hans", str(tag))
        tag = LanguageTag.fromstr("de-Latn-DE-1999")
        self.assertEqual("de-Latn-DE-1999", str(tag))

    def test_compare_case_insensitively(self):
        tag1 = LanguageTag.fromstr("de-DE")
        tag2 = LanguageTag.fromstr("de-de")
        tag3 = LanguageTag.fromstr("fr")

        self.assertTrue(tag1 == tag2)
        self.assertFalse(tag1 != tag2)
        self.assertTrue(tag2 == tag1)
        self.assertFalse(tag2 != tag1)

        self.assertTrue(tag1 != tag3)
        self.assertFalse(tag1 == tag3)
        self.assertTrue(tag2 != tag3)
        self.assertFalse(tag2 == tag3)

        self.assertTrue(tag3 != tag1)
        self.assertFalse(tag3 == tag1)
        self.assertTrue(tag3 != tag1)
        self.assertFalse(tag3 == tag1)

    def test_order_case_insensitively(self):
        tag1 = LanguageTag.fromstr("de-DE")
        tag2 = LanguageTag.fromstr("de-de")
        tag3 = LanguageTag.fromstr("en-us")
        tag4 = LanguageTag.fromstr("fr")

        self.assertLess(tag1, tag3)
        self.assertLess(tag1, tag4)
        self.assertLess(tag2, tag3)
        self.assertLess(tag2, tag4)
        self.assertLess(tag3, tag4)

        self.assertGreater(tag4, tag3)
        self.assertGreater(tag4, tag2)
        self.assertGreater(tag4, tag1)
        self.assertGreater(tag3, tag2)
        self.assertGreater(tag3, tag1)

        self.assertFalse(tag1 > tag2)
        self.assertFalse(tag2 > tag1)

        self.assertFalse(tag1 < tag2)
        self.assertFalse(tag2 < tag1)

    def test_hash_case_insensitively(self):
        tag1 = LanguageTag.fromstr("de-DE")
        tag2 = LanguageTag.fromstr("de-de")

        self.assertEqual(hash(tag1), hash(tag2))

    def test_not_equal_to_None(self):
        tag1 = LanguageTag.fromstr("de-DE")
        self.assertNotEqual(tag1, None)

    def test_dont_compare_with_None(self):
        tag1 = LanguageTag.fromstr("de-DE")
        with self.assertRaises(TypeError):
            tag1 > None  # noqa: B015
        with self.assertRaises(TypeError):
            tag1 < None  # noqa: B015
        with self.assertRaises(TypeError):
            tag1 >= None  # noqa: B015
        with self.assertRaises(TypeError):
            tag1 <= None  # noqa: B015

    def test__repr__(self):
        tag1 = LanguageTag.fromstr("de-DE")
        tag2 = LanguageTag.fromstr("fr")

        self.assertEqual("<nbxmpp.language.LanguageTag.fromstr('de-DE')>", repr(tag1))

        self.assertEqual("<nbxmpp.language.LanguageTag.fromstr('fr')>", repr(tag2))

    def test_immutable(self):
        tag = LanguageTag.fromstr("foo")
        with self.assertRaises(AttributeError):
            tag.foo = "bar"


class TestLanguageRange(unittest.TestCase):
    def test_init(self):
        with self.assertRaises(TypeError):
            LanguageRange("foo")

        with self.assertRaises(ValueError):
            LanguageRange(tag="")

        with self.assertRaises(TypeError):
            LanguageRange()

    def test_fromstr_match_str(self):
        tag = LanguageRange.fromstr("de-DE")
        self.assertEqual("de-de", tag.match_str)

    def test_fromstr_print_str(self):
        tag = LanguageRange.fromstr("de-Latn-DE-1999")
        self.assertEqual("de-Latn-DE-1999", tag.print_str)

    def test___str__(self):
        tag = LanguageRange.fromstr("zh-Hans")
        self.assertEqual("zh-Hans", str(tag))
        tag = LanguageRange.fromstr("de-Latn-DE-1999")
        self.assertEqual("de-Latn-DE-1999", str(tag))

    def test_compare_case_insensitively(self):
        tag1 = LanguageRange.fromstr("de-DE")
        tag2 = LanguageRange.fromstr("de-de")
        tag3 = LanguageRange.fromstr("fr")

        self.assertTrue(tag1 == tag2)
        self.assertFalse(tag1 != tag2)
        self.assertTrue(tag2 == tag1)
        self.assertFalse(tag2 != tag1)

        self.assertTrue(tag1 != tag3)
        self.assertFalse(tag1 == tag3)
        self.assertTrue(tag2 != tag3)
        self.assertFalse(tag2 == tag3)

        self.assertTrue(tag3 != tag1)
        self.assertFalse(tag3 == tag1)
        self.assertTrue(tag3 != tag1)
        self.assertFalse(tag3 == tag1)

    def test_hash_case_insensitively(self):
        tag1 = LanguageRange.fromstr("de-DE")
        tag2 = LanguageRange.fromstr("de-de")

        self.assertEqual(hash(tag1), hash(tag2))

    def test_not_equal_to_None(self):
        r1 = LanguageRange.fromstr("de-DE")
        self.assertNotEqual(r1, None)

    def test_wildcard(self):
        r1 = LanguageRange.fromstr("*")
        r2 = LanguageRange.fromstr("*")
        self.assertIs(r1, r2)

    def test_strip_rightmost(self):
        r = LanguageRange.fromstr("de-Latn-DE-x-foo")
        self.assertEqual(LanguageRange.fromstr("de-Latn-DE"), r.strip_rightmost())
        self.assertEqual(
            LanguageRange.fromstr("de-Latn"), r.strip_rightmost().strip_rightmost()
        )
        self.assertEqual(
            LanguageRange.fromstr("de"),
            r.strip_rightmost().strip_rightmost().strip_rightmost(),
        )
        with self.assertRaises(ValueError):
            r.strip_rightmost().strip_rightmost().strip_rightmost().strip_rightmost()

    def test_immutable(self):
        r = LanguageRange.fromstr("foo")
        with self.assertRaises(AttributeError):
            r.foo = "bar"


class TestBasicFilterLanguages(unittest.TestCase):
    def setUp(self):
        self.languages = [
            LanguageTag.fromstr("de-Latn-DE-1999"),
            LanguageTag.fromstr("de-DE"),
            LanguageTag.fromstr("de-Latn"),
            LanguageTag.fromstr("fr-CH"),
            LanguageTag.fromstr("it"),
        ]

    def test_filter(self):
        self.assertSequenceEqual(
            [
                self.languages[0],
                self.languages[1],
                self.languages[2],
            ],
            list(
                basic_filter_languages(
                    self.languages,
                    list(
                        map(
                            LanguageRange.fromstr,
                            [
                                "de",
                            ],
                        )
                    ),
                )
            ),
        )

        self.assertSequenceEqual(
            [
                self.languages[1],
            ],
            list(
                basic_filter_languages(
                    self.languages,
                    list(
                        map(
                            LanguageRange.fromstr,
                            [
                                "de-DE",
                            ],
                        )
                    ),
                )
            ),
        )

        self.assertSequenceEqual(
            [
                self.languages[0],
                self.languages[2],
            ],
            list(
                basic_filter_languages(
                    self.languages,
                    list(
                        map(
                            LanguageRange.fromstr,
                            [
                                "de-Latn",
                            ],
                        )
                    ),
                )
            ),
        )

    def test_filter_no_dupes_and_ordered(self):
        self.assertSequenceEqual(
            [
                self.languages[0],
                self.languages[2],
                self.languages[1],
            ],
            list(
                basic_filter_languages(
                    self.languages,
                    list(
                        map(
                            LanguageRange.fromstr,
                            [
                                "de-Latn",
                                "de",
                            ],
                        )
                    ),
                )
            ),
        )

    def test_filter_wildcard(self):
        self.assertSequenceEqual(
            self.languages,
            list(
                basic_filter_languages(
                    self.languages,
                    list(
                        map(
                            LanguageRange.fromstr,
                            [
                                "fr",
                                "*",
                            ],
                        )
                    ),
                )
            ),
        )


class TestLookupLanguage(unittest.TestCase):
    def setUp(self):
        self.languages = [
            LanguageTag.fromstr("de-Latn-DE-1999"),
            LanguageTag.fromstr("fr-CH"),
            LanguageTag.fromstr("it"),
        ]

    def test_match_direct(self):
        self.assertEqual(
            LanguageTag.fromstr("fr-CH"),
            lookup_language(
                self.languages,
                list(map(LanguageRange.fromstr, ["en", "fr-ch", "de-de"])),
            ),
        )

        self.assertEqual(
            LanguageTag.fromstr("it"),
            lookup_language(
                self.languages,
                list(
                    map(
                        LanguageRange.fromstr,
                        [
                            "it",
                        ],
                    )
                ),
            ),
        )

    def test_decay(self):
        self.assertEqual(
            LanguageTag.fromstr("de-Latn-DE-1999"),
            lookup_language(
                self.languages,
                list(map(LanguageRange.fromstr, ["de-de", "en-GB", "en"])),
            ),
        )

        self.assertEqual(
            LanguageTag.fromstr("fr-CH"),
            lookup_language(
                self.languages,
                list(
                    map(
                        LanguageRange.fromstr,
                        [
                            "fr-FR",
                            "de-DE",
                            "fr",
                        ],
                    )
                ),
            ),
        )

    def test_decay_skips_extension_prefixes_properly(self):
        self.assertEqual(
            LanguageTag.fromstr("de-DE"),
            lookup_language(
                list(
                    map(
                        LanguageTag.fromstr,
                        [
                            "de-DE",
                            "de-x",
                        ],
                    )
                ),
                list(
                    map(
                        LanguageRange.fromstr,
                        [
                            "de-x-foobar",
                        ],
                    )
                ),
            ),
        )


class TestLanguageMap(unittest.TestCase):
    def test_implements_mapping(self):
        mapping = LanguageMap()
        self.assertIsInstance(mapping, collections.abc.MutableMapping)

    def test_mapping_interface(self):
        key1 = LanguageTag.fromstr("de-DE")
        key2 = LanguageTag.fromstr("en-US")
        key3 = LanguageTag.fromstr("en")

        mapping = LanguageMap()

        self.assertFalse(mapping)
        self.assertEqual(0, len(mapping))

        mapping[key1] = 10

        self.assertIn(key1, mapping)
        self.assertEqual(10, mapping[key1])

        self.assertSetEqual({key1}, set(mapping))

        mapping[key2] = 20

        self.assertIn(key2, mapping)
        self.assertEqual(20, mapping[key2])

        self.assertSetEqual({key1, key2}, set(mapping))

        key2_prime = LanguageTag.fromstr("en-us")

        self.assertIn(key2_prime, mapping)
        self.assertEqual(20, mapping[key2_prime])

        self.assertNotIn(key3, mapping)

        del mapping[key1]

        self.assertNotIn(key1, mapping)

        mapping.clear()

        self.assertNotIn(key2, mapping)

    def test_lookup(self):
        key1 = LanguageTag.fromstr("de-DE")
        key2 = LanguageTag.fromstr("en-US")
        key3 = LanguageTag.fromstr("en")

        mapping = LanguageMap()

        mapping[key1] = 10
        mapping[key2] = 20
        mapping[key3] = 30

        self.assertEqual(30, mapping.lookup([LanguageRange.fromstr("en-GB")]))

    def test_values(self):
        key1 = LanguageTag.fromstr("de-DE")
        key2 = LanguageTag.fromstr("en-US")
        key3 = LanguageTag.fromstr("en")

        mapping = LanguageMap()

        mapping[key1] = 10
        mapping[key2] = 20
        mapping[key3] = 30

        self.assertSetEqual({10, 20, 30}, set(mapping.values()))

    def test_keys(self):
        key1 = LanguageTag.fromstr("de-DE")
        key2 = LanguageTag.fromstr("en-US")
        key3 = LanguageTag.fromstr("en")

        mapping = LanguageMap()

        mapping[key1] = 10
        mapping[key2] = 20
        mapping[key3] = 30

        self.assertSetEqual({key1, key2, key3}, set(mapping.keys()))

    def test_items(self):
        key1 = LanguageTag.fromstr("de-DE")
        key2 = LanguageTag.fromstr("en-US")
        key3 = LanguageTag.fromstr("en")

        mapping = LanguageMap()

        mapping[key1] = 10
        mapping[key2] = 20
        mapping[key3] = 30

        self.assertSetEqual(
            {
                (key1, 10),
                (key2, 20),
                (key3, 30),
            },
            set(mapping.items()),
        )

    def test_equality(self):
        mapping1 = LanguageMap()
        mapping1[LanguageTag.fromstr("de-de")] = 10
        mapping1[LanguageTag.fromstr("en-US")] = 20

        mapping2 = LanguageMap()
        mapping2[LanguageTag.fromstr("de-DE")] = 10
        mapping2[LanguageTag.fromstr("en-US")] = 20

        mapping3 = LanguageMap()
        mapping3[LanguageTag.fromstr("de-DE")] = 10
        mapping3[LanguageTag.fromstr("en-GB")] = 20

        self.assertEqual(mapping1, mapping2)
        self.assertFalse(mapping1 != mapping2)

        self.assertNotEqual(mapping1, mapping3)
        self.assertFalse(mapping1 == mapping3)

        self.assertNotEqual(mapping2, mapping3)
        self.assertFalse(mapping2 == mapping3)

    def test_setdefault(self):
        l: list[Any] = []
        mapping = LanguageMap()
        result = mapping.setdefault(LanguageTag.fromstr("de-de"), l)

        self.assertIs(result, l)

        result = mapping.setdefault(LanguageTag.fromstr("de-de"), [])

        self.assertIs(result, l)

    def test_lookup_returns_None_key_if_nothing_matches(self):
        mapping = LanguageMap()
        mapping[None] = "foobar"
        mapping[LanguageTag.fromstr("de")] = "Test"
        mapping[LanguageTag.fromstr("en")] = "test"

        self.assertEqual("foobar", mapping.lookup([LanguageRange.fromstr("it")]))

    def test_any_returns_only_key(self):
        m = LanguageMap()
        m[None] = "fnord"

        self.assertEqual(m.any(), "fnord")

        m = LanguageMap()
        m[LanguageTag.fromstr("de")] = "Test"

        self.assertEqual(m.any(), "Test")

    def test_any_raises_ValueError_on_empty_map(self):
        m = LanguageMap()
        with self.assertRaises(ValueError):
            m.any()

    def test_any_prefers_None(self):
        m = LanguageMap()
        m[LanguageTag.fromstr("de")] = "Test"
        m[None] = "fnord"

        self.assertEqual(m.any(), "fnord")

        m = LanguageMap()
        m[None] = "fnord"
        m[LanguageTag.fromstr("de")] = "Test"

        self.assertEqual(m.any(), "fnord")

    def test_any_returns_same_key_for_same_keyset(self):
        m = LanguageMap()
        m[LanguageTag.fromstr("de")] = "Test"
        m[LanguageTag.fromstr("fr")] = "fnord"

        self.assertEqual(m.any(), "Test")

        m = LanguageMap()
        m[LanguageTag.fromstr("fr")] = "fnord"
        m[LanguageTag.fromstr("de")] = "Test"

        self.assertEqual(m.any(), "Test")



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import logging

from nbxmpp.util import LogAdapter

if TYPE_CHECKING:
    from nbxmpp.client import Client


class BaseModule:

    _depends: dict[str, str] = {}

    def __init__(self, client: Client) -> None:
        logger_name = "nbxmpp.m.%s" % self.__class__.__name__.lower()
        self._log = LogAdapter(
            logging.getLogger(logger_name), {"context": client.log_context}
        )

    def __getattr__(self, name: str) -> Any:
        if name not in self._depends:
            raise AttributeError("Unknown method: %s" % name)

        module = self._client.get_module(self._depends[name])
        return getattr(module, name)



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from collections.abc import Callable

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import ERR_FORBIDDEN
from nbxmpp.protocol import ERR_SERVICE_UNAVAILABLE
from nbxmpp.protocol import Error
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import IqProperties
from nbxmpp.structs import SoftwareVersionResult
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class SoftwareVersion(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="iq",
                callback=self._answer_request,
                typ="get",
                priority=60,
                ns=Namespace.VERSION,
            ),
        ]

        self._name: str | None = None
        self._version: str | None = None
        self._os: str | None = None

        self._enabled = False
        self._allow_reply_func: Callable[..., bool] | None = None

    def disable(self) -> None:
        self._enabled = False

    def set_allow_reply_func(self, func: Callable[..., bool]) -> None:
        self._allow_reply_func = func

    @iq_request_task
    def request_software_version(self, jid: JID):
        _task = yield

        response = yield Iq(typ="get", to=jid, queryNS=Namespace.VERSION)
        if response.isError():
            raise StanzaError(response)

        yield _parse_info(response)

    def set_software_version(
        self, name: str, version: str, os: str | None = None
    ) -> None:
        self._name, self._version, self._os = name, version, os
        self._enabled = True

    def _answer_request(
        self, _client: Client, stanza: Iq, _properties: IqProperties
    ) -> None:
        self._log.info("Request received from %s", stanza.getFrom())
        if not self._enabled or self._name is None or self._version is None:
            self._client.send_stanza(Error(stanza, ERR_SERVICE_UNAVAILABLE))
            raise NodeProcessed

        if self._allow_reply_func is not None:
            if not self._allow_reply_func(stanza.getFrom()):
                self._client.send_stanza(Error(stanza, ERR_FORBIDDEN))
                raise NodeProcessed

        iq = stanza.buildReply("result")
        query = iq.getQuery()
        query.setTagData("name", self._name)
        query.setTagData("version", self._version)
        if self._os is not None:
            query.setTagData("os", self._os)
        self._log.info(
            "Send software version: %s %s %s", self._name, self._version, self._os
        )

        self._client.send_stanza(iq)
        raise NodeProcessed


def _parse_info(stanza: Iq) -> SoftwareVersionResult:
    try:
        name = stanza.getQueryChild("name").getData()
    except Exception:
        raise MalformedStanzaError("name node missing", stanza)

    try:
        version = stanza.getQueryChild("version").getData()
    except Exception:
        raise MalformedStanzaError("version node missing", stanza)

    os_info = stanza.getQueryChild("os")
    if os_info is not None:
        os_info = os_info.getData()

    return SoftwareVersionResult(name, version, os_info)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.const import PresenceType
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import Node
from nbxmpp.protocol import Presence
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Nickname(BaseModule):

    _depends = {"publish": "PubSub"}

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pubsub_nickname,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_nickname,
                ns=Namespace.NICK,
                priority=40,
            ),
            StanzaHandler(
                name="presence",
                callback=self._process_nickname,
                ns=Namespace.NICK,
                priority=40,
            ),
        ]

    def _process_nickname(
        self,
        _client: Client,
        stanza: Message | Presence,
        properties: MessageProperties | PresenceProperties,
    ) -> None:
        if stanza.getName() == "message":
            properties.nickname = self._parse_nickname(stanza)

        elif stanza.getName() == "presence":
            # the nickname MUST NOT be included in presence broadcasts
            # (i.e., <presence/> stanzas with no 'type' attribute or
            # of type "unavailable").
            # Usage is not recommended in MUC, but it is a workaround
            # to allow code points forbidden in resource parts in nicknames.
            if not properties.from_muc and properties.type in (
                PresenceType.AVAILABLE,
                PresenceType.UNAVAILABLE,
            ):
                return
            properties.nickname = self._parse_nickname(stanza)

    def _process_pubsub_nickname(
        self, _client: Client, _stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.NICK:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        nick = self._parse_nickname(item)
        if nick is None:
            self._log.info("Received nickname: %s - nickname removed", properties.jid)
            return

        self._log.info("Received nickname: %s - %s", properties.jid, nick)
        properties.pubsub_event = properties.pubsub_event._replace(data=nick)

    @staticmethod
    def _parse_nickname(stanza: Node) -> str | None:
        nickname = stanza.getTag("nick", namespace=Namespace.NICK)
        if nickname is None:
            return None
        return nickname.getData() or None

    @iq_request_task
    def set_nickname(self, nickname: str | None, public: bool = False):
        task = yield

        access_model = "open" if public else "presence"

        options = {
            "pubsub#persist_items": "true",
            "pubsub#access_model": access_model,
        }

        item = Node("nick", {"xmlns": Namespace.NICK})
        if nickname is not None:
            item.addData(nickname)

        result = yield self.publish(
            Namespace.NICK,
            item,
            id_="current",
            options=options,
            force_node_options=True,
        )

        yield finalize(task, result)

    @iq_request_task
    def set_access_model(self, public: bool):
        task = yield

        access_model = "open" if public else "presence"

        result = yield self._client.get_module("PubSub").set_access_model(
            Namespace.NICK, access_model
        )

        yield finalize(task, result)



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.const import LOCATION_DATA
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import Node
from nbxmpp.structs import LocationData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Location(BaseModule):

    _depends = {"publish": "PubSub"}

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pubsub_location,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            ),
        ]

    def _process_pubsub_location(
        self, _client: Client, _stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.LOCATION:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        location_node = item.getTag("geoloc", namespace=Namespace.LOCATION)
        if not location_node.getChildren():
            self._log.info("Received location: %s - no location set", properties.jid)
            return

        location_dict = {}
        for node in LOCATION_DATA:
            location_dict[node] = location_node.getTagData(node)
        data = LocationData(**location_dict)
        pubsub_event = properties.pubsub_event._replace(data=data)
        self._log.info("Received location: %s - %s", properties.jid, data)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def set_location(self, data: LocationData | None):
        task = yield

        item = Node("geoloc", {"xmlns": Namespace.LOCATION})
        if data is not None:
            data = data._asdict()
            for tag, value in data.items():
                if value is not None:
                    item.addChild(tag, payload=value)

        result = yield self.publish(Namespace.LOCATION, item, id_="current")

        yield finalize(task, result)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any

import logging

from nbxmpp.modules.dataforms import MultipleDataForm
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Protocol
from nbxmpp.structs import RegisterData


def is_error(error: Any) -> bool:
    return isinstance(error, BaseError)


class BaseError(Exception):
    def __init__(self, is_fatal: bool = False) -> None:
        self.is_fatal = is_fatal
        self.text = ""

    def __str__(self) -> str:
        return self.text

    def get_text(self, _pref_lang: str | None = None) -> str:
        return self.text


class StanzaError(BaseError):

    log_level = logging.INFO
    app_namespace = None

    def __init__(self, stanza: Protocol) -> None:
        BaseError.__init__(self)
        self.stanza = stanza
        self._stanza_name = stanza.getName()
        self._error_node = stanza.getTag("error")
        self.condition: str | None = stanza.getError()
        self.condition_data = self._error_node.getTagData(self.condition)
        self.app_condition = self._get_app_condition()
        self.type = stanza.getErrorType()
        self.jid = stanza.getFrom()
        self.id = stanza.getID()
        self._text: dict[str, str] = {}

        text_elements = self._error_node.getTags("text", namespace=Namespace.STANZAS)
        for element in text_elements:
            lang = element.getXmlLang()
            text = element.getData()
            self._text[lang] = text

    def _get_app_condition(self) -> str | None:
        if self.app_namespace is None:
            return None

        for node in self._error_node.getChildren():
            if node.getNamespace() == self.app_namespace:
                return node.getName()
        return None

    def get_text(self, pref_lang: str | None = None) -> str:
        if pref_lang is not None:
            text = self._text.get(pref_lang)
            if text is not None:
                return text

        if self._text:
            text = self._text.get("en")
            if text is not None:
                return text

            text = self._text.get(None)
            if text is not None:
                return text
            return self._text.popitem()[1]
        return ""

    def set_text(self, lang: str, text: str) -> None:
        self._text[lang] = text

    def __str__(self) -> str:
        condition = self.condition
        if self.app_condition is not None:
            condition = "%s (%s)" % (self.condition, self.app_condition)
        text = self.get_text("en") or ""
        if text:
            text = " - %s" % text
        return "Error from %s: %s%s" % (self.jid, condition, text)


class PubSubStanzaError(StanzaError):

    app_namespace = Namespace.PUBSUB_ERROR


class HTTPUploadStanzaError(StanzaError):

    app_namespace = Namespace.HTTPUPLOAD_0

    def get_max_file_size(self) -> float | None:
        if self.app_condition != "file-too-large":
            return None

        node = self._error_node.getTag(self.app_condition)
        try:
            return float(node.getTagData("max-file-size"))
        except Exception:
            return None

    def get_retry_date(self) -> str | None:
        if self.app_condition != "retry":
            return None
        return self._error_node.getTagAttr("stamp")


class MalformedStanzaError(BaseError):

    log_level = logging.WARNING

    def __init__(self, text: str, stanza: Protocol, is_fatal: bool = True) -> None:
        BaseError.__init__(self, is_fatal=is_fatal)
        self.stanza = stanza
        self.text = str(text)


class CancelledError(BaseError):

    log_level = logging.INFO

    def __init__(self) -> None:
        BaseError.__init__(self, is_fatal=True)
        self.text = "Task has been cancelled"


class TimeoutStanzaError(BaseError):

    log_level = logging.INFO

    def __init__(self) -> None:
        BaseError.__init__(self)
        self.text = "Timeout reached"


class RegisterStanzaError(StanzaError):
    def __init__(self, stanza: Protocol, data: RegisterData) -> None:
        StanzaError.__init__(self, stanza)
        self._data = data

    def get_data(self) -> RegisterData:
        return self._data


class ChangePasswordStanzaError(StanzaError):
    def __init__(
        self, stanza: Protocol, form: SimpleDataForm | MultipleDataForm
    ) -> None:
        StanzaError.__init__(self, stanza)
        self._form = form

    def get_form(self) -> SimpleDataForm | MultipleDataForm:
        return self._form



from typing import Literal

from nbxmpp.const import ConnectionProtocol
from nbxmpp.const import ConnectionType

BlockingReportValues = Literal["spam", "abuse"]
CustomHostT = tuple[str, ConnectionProtocol, ConnectionType]



from test.lib.util import StanzaHandlerTest

from nbxmpp.modules.user_avatar import AvatarInfo
from nbxmpp.modules.user_avatar import AvatarMetaData
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import PubSubEventData
from nbxmpp.structs import StanzaHandler


class AvatarTest(StanzaHandlerTest):

    def test_avatar_parsing(self):
        def _on_message(_con, _stanza, properties):
            info = AvatarInfo(
                bytes="12345",
                height="64",
                width="64",
                id="111f4b3c50d7b0df729d299bc6f8e9ef9066971f",
                type="image/png",
                url="http://avatars.example.org/happy.gif",
            )

            metadata = AvatarMetaData(infos=[info])
            pubsub_event = PubSubEventData(
                node="urn:xmpp:avatar:metadata",
                id="111f4b3c50d7b0df729d299bc6f8e9ef9066971f",
                item=None,
                data=metadata,
                deleted=False,
                retracted=False,
                purged=False,
            )

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = """
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='urn:xmpp:avatar:metadata'>
                        <item id='111f4b3c50d7b0df729d299bc6f8e9ef9066971f'>
                            <metadata xmlns='urn:xmpp:avatar:metadata'>
                                <info bytes='12345'
                                      height='64'
                                      id='111f4b3c50d7b0df729d299bc6f8e9ef9066971f'
                                      type='image/png'
                                      width='64'
                                      url='http://avatars.example.org/happy.gif'/>
                            </metadata>
                        </item>
                    </items>
                </event>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message", callback=_on_message, ns=Namespace.PUBSUB_EVENT
            )
        )

        self.dispatcher.process_data(event)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import Node
from nbxmpp.protocol import validate_resourcepart
from nbxmpp.structs import BookmarkData
from nbxmpp.util import from_xs_boolean
from nbxmpp.util import to_xs_boolean


def parse_nickname(nick: str | None) -> str | None:
    if nick is None:
        return None

    try:
        return validate_resourcepart(nick)
    except Exception:
        return None


def parse_autojoin(autojoin: str | None) -> bool:
    if autojoin is None:
        return False

    try:
        return from_xs_boolean(autojoin)
    except ValueError:
        return False


def parse_bookmark(item: Node) -> BookmarkData:
    conference = item.getTag("conference", namespace=Namespace.BOOKMARKS_1)
    if conference is None:
        raise MalformedStanzaError("conference node missing", item)

    try:
        jid = JID.from_string(item.getAttr("id"))
    except Exception as error:
        raise MalformedStanzaError("invalid jid: %s" % error, item)

    if jid.localpart is None or jid.resource is not None:
        raise MalformedStanzaError("invalid jid", item)

    autojoin = parse_autojoin(conference.getAttr("autojoin"))
    nick = parse_nickname(conference.getTagData("nick"))
    name = conference.getAttr("name") or None
    password = conference.getTagData("password") or None
    extensions = conference.getTag("extensions")

    return BookmarkData(
        jid=jid,
        name=name,
        nick=nick,
        autojoin=autojoin,
        password=password,
        extensions=extensions,
    )


def parse_bookmarks(item: Node, log: logging.Logger) -> list[BookmarkData]:
    storage_node = item.getTag("storage", namespace=Namespace.BOOKMARKS)
    if storage_node is None:
        raise MalformedStanzaError("storage node missing", item)

    return parse_storage_node(storage_node, log)


def parse_private_bookmarks(response: Node, log: logging.Logger) -> list[BookmarkData]:
    query = response.getQuery()
    storage_node = query.getTag("storage", namespace=Namespace.BOOKMARKS)
    if storage_node is None:
        raise MalformedStanzaError("storage node missing", response)

    return parse_storage_node(storage_node, log)


def parse_storage_node(storage: Node, log: logging.Logger) -> list[BookmarkData]:
    bookmarks: list[BookmarkData] = []
    confs = storage.getTags("conference")
    for conf in confs:
        try:
            jid = JID.from_string(conf.getAttr("jid"))
        except Exception:
            log.warning("invalid jid: %s", conf)
            continue

        if jid.localpart is None or jid.resource is not None:
            log.warning("invalid jid: %s", conf)
            continue

        autojoin = parse_autojoin(conf.getAttr("autojoin"))
        nick = parse_nickname(conf.getTagData("nick"))
        name = conf.getAttr("name") or None
        password = conf.getTagData("password") or None

        bookmark = BookmarkData(
            jid=jid, name=name, autojoin=autojoin, password=password, nick=nick
        )
        bookmarks.append(bookmark)

    return bookmarks


def build_conference_node(bookmark: BookmarkData):
    attrs = {"xmlns": Namespace.BOOKMARKS_1}
    if bookmark.autojoin:
        attrs["autojoin"] = "true"
    if bookmark.name:
        attrs["name"] = bookmark.name
    conference = Node(tag="conference", attrs=attrs)
    if bookmark.nick:
        conference.setTagData("nick", bookmark.nick)
    if bookmark.password:
        conference.setTagData("password", bookmark.password)
    if bookmark.extensions is not None:
        conference.addChild(node=bookmark.extensions)
    return conference


def build_storage_node(bookmarks: list[BookmarkData]):
    storage_node = Node(tag="storage", attrs={"xmlns": Namespace.BOOKMARKS})
    for bookmark in bookmarks:
        conf_node = storage_node.addChild(name="conference")
        conf_node.setAttr("jid", bookmark.jid)
        conf_node.setAttr("autojoin", to_xs_boolean(bookmark.autojoin))
        if bookmark.name:
            conf_node.setAttr("name", bookmark.name)
        if bookmark.nick:
            conf_node.setTagData("nick", bookmark.nick)
        if bookmark.password:
            conf_node.setTagData("password", bookmark.password)
    return storage_node


def get_private_request() -> Iq:
    iq = Iq(typ="get")
    query = iq.addChild(name="query", namespace=Namespace.PRIVATE)
    query.addChild(name="storage", namespace=Namespace.BOOKMARKS)
    return iq



from test.lib.util import StanzaHandlerTest

from nbxmpp.modules.ogp import OpenGraph
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import OpenGraphData


class MockOpenGraph(OpenGraph):
    def __init__(self) -> None:
        pass


class OpenGraphTest(StanzaHandlerTest):
    def test_parse_ogp(self) -> None:
        msg = Message(node=_DATA)
        self._parse_and_assert_values(msg)

    def test_parse_wikipedia(self) -> None:
        msg = Message(
            node="""
        <message xmlns="jabber:client" id="ef1bee13-ada9-4de8-8d1d-5ccce250116d" from="gajim-unnamed-chat@rooms.slidge.im/nicoco" type="groupchat" xml:lang="en" to="test@slidge.im/gajim.M7JMH3QS">
          <markable xmlns="urn:xmpp:chat-markers:0" />
          <Description xmlns="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:ns1="http://www.w3.org/1999/02/22-rdf-syntax-ns#" ns1:about="https://fr.wikipedia.org">
            <title xmlns="https://ogp.me/ns#">Wikipédia, l'encyclopédie libre</title>
            <url xmlns="https://ogp.me/ns#">https://fr.wikipedia.org/wiki/Wikip%C3%A9dia:Accueil_principal</url>
        </Description>
          <body>https://fr.wikipedia.org</body>
          <occupant-id xmlns="urn:xmpp:occupant-id:0" id="5QLCTSehDZCqSoQznCE5puX3OfkZENjjCsFb35m6Gyw=" />
          <stanza-id id="068e5132-780b-75c4-82cb-cbed329a08a5" xmlns="urn:xmpp:sid:0" by="gajim-unnamed-chat@rooms.slidge.im" />
        </message>
        """
        )
        props = self._parse(msg)
        ogp = props["https://fr.wikipedia.org"]
        self.assertEqual(ogp.title, "Wikipédia, l'encyclopédie libre")
        self.assertEqual(
            ogp.url, "https://fr.wikipedia.org/wiki/Wikip%C3%A9dia:Accueil_principal"
        )

    def test_parse_youtube(self) -> None:
        msg = Message(
            node="""
        <message xmlns="jabber:client" id="200bd533-99c6-4d0a-818d-b3ed9289d405" from="gajim-unnamed-chat@rooms.slidge.im/nicoco" type="groupchat" xml:lang="en" to="test@slidge.im/gajim.M7JMH3QS">
          <markable xmlns="urn:xmpp:chat-markers:0" />
          <Description xmlns="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:ns1="http://www.w3.org/1999/02/22-rdf-syntax-ns#" ns1:about="https://www.youtube.com/watch?v=45V5qdjhbGI&amp;t=1">
            <title xmlns="https://ogp.me/ns#">Pierre Lapointe - Je déteste ma vie (Paris tristesse) - YouTube</title>
            <description xmlns="https://ogp.me/ns#">Paris tristesse en vente maintenant: http://smarturl.it/paristristesseRéalisation | montage : Pascal Grandmaison et Philippe CraigDirection photo: Pascal Gra...</description>
            <url xmlns="https://ogp.me/ns#">https://www.youtube.com/watch?v=45V5qdjhbGI&amp;t=1</url>
            <site_name xmlns="https://ogp.me/ns#">YouTube</site_name>
        </Description>
          <body>https://www.youtube.com/watch?v=45V5qdjhbGI&amp;t=1</body>
          <occupant-id xmlns="urn:xmpp:occupant-id:0" id="5QLCTSehDZCqSoQznCE5puX3OfkZENjjCsFb35m6Gyw=" />
          <stanza-id id="068e51c3-04f6-7560-96ae-45c83f4bce1c" xmlns="urn:xmpp:sid:0" by="gajim-unnamed-chat@rooms.slidge.im" />
        </message>
        """
        )
        props = self._parse(msg)
        ogp = props["https://www.youtube.com/watch?v=45V5qdjhbGI&t=1"]
        self.assertEqual(
            ogp.title, "Pierre Lapointe - Je déteste ma vie (Paris tristesse) - YouTube"
        )
        self.assertEqual(
            ogp.description,
            "Paris tristesse en vente maintenant: http://smarturl.it/paristristesseRéalisation | montage : Pascal Grandmaison et Philippe CraigDirection photo: Pascal Gra...",
        )
        self.assertEqual(ogp.site_name, "YouTube")

    def test_set_ogp(self) -> None:
        msg1 = Message(node=_DATA)
        msg2 = Message(to=msg1.getTo(), body=msg1.getBody())
        msg2.addOpenGraph(
            "https://the.link.example.com/what-was-linked-to",
            OpenGraphData(
                title="Page Title",
                url="Canonical URL",
                image="https://link.to.example.com/image.png",
                type="website",
                site_name="Some Website",
                description="Page Description",
            ),
        )
        self._parse_and_assert_values(msg2)

    def _parse(self, msg: Message) -> dict[str, OpenGraphData]:
        properties = MessageProperties("whatever@whatever.com")
        parser = MockOpenGraph()
        parser._process_message_opengraph(None, msg, properties)
        self.assertEqual(len(properties.open_graph), 1)

        return properties.open_graph

    def _parse_and_assert_values(self, msg: Message) -> None:
        props = self._parse(msg)
        ogp = props["https://the.link.example.com/what-was-linked-to"]
        self.assertEqual(ogp.title, "Page Title")
        self.assertEqual(ogp.url, "Canonical URL")
        self.assertEqual(ogp.image, "https://link.to.example.com/image.png")
        self.assertEqual(ogp.type, "website")
        self.assertEqual(ogp.site_name, "Some Website")
        self.assertEqual(ogp.description, "Page Description")


_DATA = """
<message xmlns="jabber:client"
         to="whoever@example.com">
    <body>I wanted to mention https://the.link.example.com/what-was-linked-to</body>
    <rdf:Description xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
                     xmlns:og="https://ogp.me/ns#"
                     rdf:about="https://the.link.example.com/what-was-linked-to">
        <og:title>Page Title</og:title>
        <og:description>Page Description</og:description>
        <og:url>Canonical URL</og:url>
        <og:image>https://link.to.example.com/image.png</og:image>
        <og:type>website</og:type>
        <og:site_name>Some Website</og:site_name>
    </rdf:Description>
</message>
"""



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any

import logging

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Soup

from nbxmpp.connection import Connection
from nbxmpp.const import TCPState
from nbxmpp.http import DEFAULT_USER_AGENT
from nbxmpp.protocol import Protocol
from nbxmpp.util import convert_tls_error_flags
from nbxmpp.util import get_websocket_close_string

log = logging.getLogger("nbxmpp.websocket")


class WebsocketConnection(Connection):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        Connection.__init__(self, *args, **kwargs)

        self._session = Soup.Session()
        self._session.set_user_agent(DEFAULT_USER_AGENT)

        if self._log.getEffectiveLevel() == logging.INFO:
            self._session.add_feature(Soup.Logger.new(Soup.LoggerLogLevel.BODY))

        self._websocket: Soup.WebsocketConnection | None = None
        self._cancellable = Gio.Cancellable()

        self._input_closed = False
        self._output_closed = False

    def connect(self) -> None:
        assert self._address is not None
        self._log.info("Try to connect to %s", self._address.uri)

        self.state = TCPState.CONNECTING

        message = Soup.Message.new("GET", self._address.uri)
        assert message is not None
        message.connect("accept-certificate", self._check_certificate)
        message.connect("notify::tls-peer-certificate", self._on_certificate_set)
        message.connect("network-event", self._on_network_event)
        message.set_flags(Soup.MessageFlags.NO_REDIRECT)
        assert self._session is not None
        self._session.websocket_connect_async(
            message,
            None,
            ["xmpp"],
            GLib.PRIORITY_DEFAULT,
            self._cancellable,
            self._on_connect,
            None,
        )

    def _on_connect(
        self, session: Soup.Session, result: Gio.AsyncResult, _user_data: None
    ) -> None:
        # TODO: check if protocol 'xmpp' is set
        try:
            self._websocket = session.websocket_connect_finish(result)
        except GLib.Error as error:
            quark = GLib.quark_try_string("g-io-error-quark")
            if error.matches(quark, Gio.IOErrorEnum.CANCELLED):
                self._finalize("disconnected")
                return

            self._log.info("Connection Error: %s", error)
            self._finalize("connection-failed")
            return

        self._websocket.set_keepalive_interval(5)
        self._websocket.set_max_incoming_payload_size(1048576)
        self._websocket.connect("message", self._on_websocket_message)
        self._websocket.connect("closed", self._on_websocket_closed)
        self._websocket.connect("closing", self._on_websocket_closing)
        self._websocket.connect("error", self._on_websocket_error)
        self._websocket.connect("pong", self._on_websocket_pong)

        self.state = TCPState.CONNECTED
        self.notify("connected")

    def start_tls_negotiation(self) -> None:
        # Soup.Session does this automatically
        raise NotImplementedError

    def _check_certificate(
        self,
        _message: Soup.Message,
        certificate: Gio.TlsCertificate,
        errors: Gio.TlsCertificateFlags,
    ) -> bool:

        self._peer_certificate = certificate
        self._peer_certificate_errors = convert_tls_error_flags(errors)

        if self._accept_certificate():
            return True

        self.notify("bad-certificate")
        self._cancellable.cancel()
        return False

    def _on_network_event(
        self,
        _message: Soup.Message,
        event: Gio.SocketClientEvent,
        connection: Gio.TlsConnection,
    ) -> None:

        if event == Gio.SocketClientEvent.TLS_HANDSHAKED:
            self._tls_con = connection

    def _on_certificate_set(self, message: Soup.Message, _param: Any) -> None:
        if self._peer_certificate is not None:
            return

        # If the cert has errors _check_certificate() will set the cert.
        self._peer_certificate = message.props.tls_peer_certificate
        self._peer_certificate_errors = convert_tls_error_flags(
            message.props.tls_peer_certificate_errors
        )

    def _on_websocket_message(
        self, _websocket: Soup.WebsocketConnection, _type: int, message: GLib.Bytes
    ) -> None:
        data = message.get_data().decode()
        self._log_stanza(data)

        if self._input_closed:
            self._log.warning("Received data after stream closed")
            return

        self.notify("data-received", data)

    def _on_websocket_pong(
        self, _websocket: Soup.WebsocketConnection, _message: GLib.Bytes
    ) -> None:
        self._log.info("Pong received")

    def _on_websocket_closed(self, websocket: Soup.WebsocketConnection) -> None:
        self._log.info("Closed %s", get_websocket_close_string(websocket))
        self._finalize("disconnected")

    def _on_websocket_closing(self, _websocket: Soup.WebsocketConnection) -> None:
        self._log.info("Closing")

    def _on_websocket_error(
        self, _websocket: Soup.WebsocketConnection, error: GLib.Error
    ) -> None:
        self._log.error(error)
        if self._state not in (TCPState.DISCONNECTED, TCPState.DISCONNECTING):
            self._finalize("disconnected")

    def send(self, stanza: Protocol, now: bool = False) -> None:
        if self._state in (TCPState.DISCONNECTED, TCPState.DISCONNECTING):
            self._log.warning("send() not possible in state: %s", self._state)
            return

        data = str(stanza)
        self._websocket.send_text(data)
        self._log_stanza(data, received=False)
        self.notify("data-sent", stanza)

    def disconnect(self) -> None:
        if self._state == TCPState.CONNECTING:
            self.state = TCPState.DISCONNECTING
            self._cancellable.cancel()
            return

        if self._state in (TCPState.DISCONNECTED, TCPState.DISCONNECTING):
            self._log.warning("Called disconnect on state: %s", self._state)
            return

        self._websocket.close(Soup.WebsocketCloseCode.NORMAL, None)
        self.state = TCPState.DISCONNECTING

    def _check_for_shutdown(self) -> None:
        if self._input_closed and self._output_closed:
            self._websocket.close(Soup.WebsocketCloseCode.NORMAL, None)

    def shutdown_input(self) -> None:
        self._log.info("Shutdown input")
        self._input_closed = True
        self._check_for_shutdown()

    def shutdown_output(self) -> None:
        self.state = TCPState.DISCONNECTING
        self._log.info("Shutdown output")
        self._output_closed = True

    def _finalize(self, signal_name: str) -> None:
        if self._websocket is not None:
            self._websocket.disconnect_by_func(self._on_websocket_message)
            self._websocket.disconnect_by_func(self._on_websocket_closed)
            self._websocket.disconnect_by_func(self._on_websocket_closing)
            self._websocket.disconnect_by_func(self._on_websocket_error)
            self._websocket.disconnect_by_func(self._on_websocket_pong)

        self._input_closed = True
        self._output_closed = True
        self.state = TCPState.DISCONNECTED
        self.notify(signal_name)
        self.destroy()

    def destroy(self) -> None:
        super().destroy()
        self._session.abort()
        self._session = None
        self._websocket = None



from test.lib.util import StanzaHandlerTest

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import BookmarkData
from nbxmpp.structs import PubSubEventData
from nbxmpp.structs import StanzaHandler


class BookmarkTest(StanzaHandlerTest):

    def test_bookmark_1_parsing(self):
        def _on_message(_con, _stanza, properties):

            data = [
                BookmarkData(
                    jid=JID.from_string("theplay@conference.shakespeare.lit"),
                    name="The Play's the Thing",
                    autojoin=True,
                    password="pass",  # noqa: S106
                    nick="JC",
                ),
                BookmarkData(
                    jid=JID.from_string("second@conference.shakespeare.lit"),
                    name="Second room",
                    autojoin=False,
                    password=None,
                    nick=None,
                ),
            ]

            pubsub_event = PubSubEventData(
                node="storage:bookmarks",
                id="current",
                item=None,
                data=data,
                deleted=False,
                retracted=False,
                purged=False,
            )

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = """
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='storage:bookmarks'>
                        <item id='current'>
                            <storage xmlns='storage:bookmarks'>
                                <conference name='The Play&apos;s the Thing'
                                            autojoin='true'
                                            jid='theplay@conference.shakespeare.lit'>
                                    <password>pass</password>
                                    <nick>JC</nick>
                                </conference>
                                <conference name='Second room'
                                            autojoin='0'
                                            jid='second@conference.shakespeare.lit'>
                                </conference>
                            </storage>
                        </item>
                    </items>
                </event>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message", callback=_on_message, ns=Namespace.PUBSUB_EVENT
            )
        )

        self.dispatcher.process_data(event)

    def test_bookmark_2_parsing(self):
        def _on_message(_con, _stanza, properties):

            data = BookmarkData(
                jid=JID.from_string("theplay@conference.shakespeare.lit"),
                name="The Play's the Thing",
                autojoin=True,
                password=None,
                nick="JC",
            )

            pubsub_event = PubSubEventData(
                node="urn:xmpp:bookmarks:0",
                id="theplay@conference.shakespeare.lit",
                item=None,
                data=data,
                deleted=False,
                retracted=False,
                purged=False,
            )

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = """
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='urn:xmpp:bookmarks:0'>
                        <item id='theplay@conference.shakespeare.lit'>
                            <conference xmlns='urn:xmpp:bookmarks:0'
                                        name='The Play&apos;s the Thing'
                                        autojoin='1'>
                                <nick>JC</nick>
                            </conference>
                        </item>
                    </items>
                </event>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message", callback=_on_message, ns=Namespace.PUBSUB_EVENT
            )
        )

        self.dispatcher.process_data(event)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from enum import Enum
from enum import IntEnum
from functools import total_ordering

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Soup
from packaging.version import Version

GLIB_VERSION = Version(
    f"{GLib.MAJOR_VERSION}.{GLib.MINOR_VERSION}.{GLib.MICRO_VERSION}"
)


class IqType(Enum):
    GET = "get"
    SET = "set"
    RESULT = "result"
    ERROR = "error"

    @property
    def is_get(self) -> bool:
        return self == IqType.GET

    @property
    def is_set(self) -> bool:
        return self == IqType.SET

    @property
    def is_result(self) -> bool:
        return self == IqType.RESULT

    @property
    def is_error(self) -> bool:
        return self == IqType.ERROR


class MessageType(Enum):
    NORMAL = "normal"
    CHAT = "chat"
    GROUPCHAT = "groupchat"
    HEADLINE = "headline"
    ERROR = "error"

    @property
    def is_normal(self) -> bool:
        return self == MessageType.NORMAL

    @property
    def is_chat(self) -> bool:
        return self == MessageType.CHAT

    @property
    def is_groupchat(self) -> bool:
        return self == MessageType.GROUPCHAT

    @property
    def is_headline(self) -> bool:
        return self == MessageType.HEADLINE

    @property
    def is_error(self) -> bool:
        return self == MessageType.ERROR


class PresenceType(Enum):
    PROBE = "probe"
    SUBSCRIBE = "subscribe"
    SUBSCRIBED = "subscribed"
    AVAILABLE = None
    UNAVAILABLE = "unavailable"
    UNSUBSCRIBE = "unsubscribe"
    UNSUBSCRIBED = "unsubscribed"
    ERROR = "error"

    @property
    def is_available(self) -> bool:
        return self == PresenceType.AVAILABLE

    @property
    def is_unavailable(self) -> bool:
        return self == PresenceType.UNAVAILABLE

    @property
    def is_error(self) -> bool:
        return self == PresenceType.ERROR

    @property
    def is_probe(self) -> bool:
        return self == PresenceType.PROBE

    @property
    def is_unsubscribe(self) -> bool:
        return self == PresenceType.UNSUBSCRIBE

    @property
    def is_unsubscribed(self) -> bool:
        return self == PresenceType.UNSUBSCRIBED

    @property
    def is_subscribe(self) -> bool:
        return self == PresenceType.SUBSCRIBE

    @property
    def is_subscribed(self) -> bool:
        return self == PresenceType.SUBSCRIBED


@total_ordering
class PresenceShow(Enum):
    ONLINE = "online"
    CHAT = "chat"
    AWAY = "away"
    XA = "xa"
    DND = "dnd"

    @property
    def is_online(self) -> bool:
        return self == PresenceShow.ONLINE

    @property
    def is_chat(self) -> bool:
        return self == PresenceShow.CHAT

    @property
    def is_away(self) -> bool:
        return self == PresenceShow.AWAY

    @property
    def is_xa(self) -> bool:
        return self == PresenceShow.XA

    @property
    def is_dnd(self) -> bool:
        return self == PresenceShow.DND

    def __lt__(self, other: PresenceShow) -> bool:
        try:
            w1 = self._WEIGHTS[self]
            w2 = self._WEIGHTS[other]
        except KeyError:
            return NotImplemented
        return w1 < w2


PresenceShow._WEIGHTS = {
    PresenceShow.DND: 2,
    PresenceShow.CHAT: 1,
    PresenceShow.ONLINE: 0,
    PresenceShow.AWAY: -1,
    PresenceShow.XA: -2,
}


@total_ordering
class Chatstate(Enum):
    COMPOSING = "composing"
    PAUSED = "paused"
    ACTIVE = "active"
    INACTIVE = "inactive"
    GONE = "gone"

    @property
    def is_composing(self) -> bool:
        return self == Chatstate.COMPOSING

    @property
    def is_paused(self) -> bool:
        return self == Chatstate.PAUSED

    @property
    def is_active(self) -> bool:
        return self == Chatstate.ACTIVE

    @property
    def is_inactive(self) -> bool:
        return self == Chatstate.INACTIVE

    @property
    def is_gone(self) -> bool:
        return self == Chatstate.GONE

    def __lt__(self, other: Chatstate) -> bool:
        try:
            w1 = self._WEIGHTS[self]
            w2 = self._WEIGHTS[other]
        except KeyError:
            return NotImplemented
        return w1 < w2


Chatstate._WEIGHTS = {
    Chatstate.COMPOSING: 0,
    Chatstate.PAUSED: 1,
    Chatstate.ACTIVE: 2,
    Chatstate.INACTIVE: 3,
    Chatstate.GONE: 4,
}


class StatusCode(Enum):
    NON_ANONYMOUS = "100"
    AFFILIATION_CHANGE = "101"
    SHOWING_UNAVAILABLE = "102"
    NOT_SHOWING_UNAVAILABLE = "103"
    CONFIG_NON_PRIVACY_RELATED = "104"
    SELF = "110"
    CONFIG_ROOM_LOGGING = "170"
    CONFIG_NO_ROOM_LOGGING = "171"
    CONFIG_NON_ANONYMOUS = "172"
    CONFIG_SEMI_ANONYMOUS = "173"
    CONFIG_FULL_ANONYMOUS = "174"
    CREATED = "201"
    NICKNAME_MODIFIED = "210"
    REMOVED_BANNED = "301"
    NICKNAME_CHANGE = "303"
    REMOVED_KICKED = "307"
    REMOVED_AFFILIATION_CHANGE = "321"
    REMOVED_NONMEMBER_IN_MEMBERS_ONLY = "322"
    REMOVED_SERVICE_SHUTDOWN = "332"
    REMOVED_ERROR = "333"


class InviteType(Enum):
    MEDIATED = "mediated"
    DIRECT = "direct"


class AvatarState(Enum):
    IGNORE = "ignore"
    NOT_READY = "not ready"
    EMPTY = "empty"
    ADVERTISED = "advertised"


@total_ordering
class Affiliation(Enum):
    OWNER = "owner"
    ADMIN = "admin"
    MEMBER = "member"
    OUTCAST = "outcast"
    NONE = "none"

    @property
    def is_owner(self) -> bool:
        return self == Affiliation.OWNER

    @property
    def is_admin(self) -> bool:
        return self == Affiliation.ADMIN

    @property
    def is_member(self) -> bool:
        return self == Affiliation.MEMBER

    @property
    def is_outcast(self) -> bool:
        return self == Affiliation.OUTCAST

    @property
    def is_none(self) -> bool:
        return self == Affiliation.NONE

    def __lt__(self, other: Affiliation) -> bool:
        try:
            w1 = self._WEIGHTS[self]
            w2 = self._WEIGHTS[other]
        except KeyError:
            return NotImplemented
        return w1 < w2


Affiliation._WEIGHTS = {
    Affiliation.OWNER: 4,
    Affiliation.ADMIN: 3,
    Affiliation.MEMBER: 2,
    Affiliation.NONE: 1,
    Affiliation.OUTCAST: 0,
}


@total_ordering
class Role(Enum):
    MODERATOR = "moderator"
    PARTICIPANT = "participant"
    VISITOR = "visitor"
    NONE = "none"

    @property
    def is_moderator(self) -> bool:
        return self == Role.MODERATOR

    @property
    def is_participant(self) -> bool:
        return self == Role.PARTICIPANT

    @property
    def is_visitor(self) -> bool:
        return self == Role.VISITOR

    @property
    def is_none(self) -> bool:
        return self == Role.NONE

    def __lt__(self, other: Role) -> bool:
        try:
            w1 = self._WEIGHTS[self]
            w2 = self._WEIGHTS[other]
        except KeyError:
            return NotImplemented
        return w1 < w2


Role._WEIGHTS = {
    Role.MODERATOR: 3,
    Role.PARTICIPANT: 2,
    Role.VISITOR: 1,
    Role.NONE: 0,
}


class AnonymityMode(Enum):
    UNKNOWN = None
    SEMI = "semi"
    NONE = "none"


class AdHocStatus(Enum):
    EXECUTING = "executing"
    COMPLETED = "completed"
    CANCELED = "canceled"


class AdHocAction(Enum):
    EXECUTE = "execute"
    CANCEL = "cancel"
    PREV = "prev"
    NEXT = "next"
    COMPLETE = "complete"


class AdHocNoteType(Enum):
    INFO = "info"
    WARN = "warn"
    ERROR = "error"


class ConnectionType(Enum):
    DIRECT_TLS = "DIRECT TLS"
    START_TLS = "START TLS"
    PLAIN = "PLAIN"

    @property
    def is_direct_tls(self) -> bool:
        return self == ConnectionType.DIRECT_TLS

    @property
    def is_start_tls(self) -> bool:
        return self == ConnectionType.START_TLS

    @property
    def is_plain(self) -> bool:
        return self == ConnectionType.PLAIN


class ConnectionProtocol(IntEnum):
    TCP = 0
    WEBSOCKET = 1


class StreamState(Enum):
    RESOLVE = "resolve"
    RESOLVED = "resolved"
    CONNECTING = "connecting"
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    DISCONNECTING = "disconnecting"
    STREAM_START = "stream start"
    WAIT_FOR_STREAM_START = "wait for stream start"
    WAIT_FOR_FEATURES = "wait for features"
    WAIT_FOR_TLS_PROCEED = "wait for tls proceed"
    TLS_START_SUCCESSFUL = "tls start successful"
    PROCEED_WITH_AUTH = "proceed with auth"
    AUTH_SUCCESSFUL = "auth successful"
    AUTH_FAILED = "auth failed"
    WAIT_FOR_RESUMED = "wait for resumed"
    RESUME_FAILED = "resume failed"
    RESUME_SUCCESSFUL = "resume successful"
    PROCEED_WITH_BIND = "proceed with bind"
    BIND_SUCCESSFUL = "bind successful"
    WAIT_FOR_BIND = "wait for bind"
    WAIT_FOR_SESSION = "wait for session"
    ACTIVE = "active"


class StreamError(Enum):
    PARSING = 0
    CONNECTION_FAILED = 1
    SESSION = 2
    BIND = 3
    TLS = 4
    BAD_CERTIFICATE = 5
    STREAM = 6
    SASL = 7
    REGISTER = 8
    END = 9


class TCPState(Enum):
    DISCONNECTED = "disconnected"
    DISCONNECTING = "disconnecting"
    CONNECTING = "connecting"
    CONNECTED = "connected"


class Mode(IntEnum):
    CLIENT = 0
    REGISTER = 1
    LOGIN_TEST = 2
    ANONYMOUS_TEST = 3

    @property
    def is_client(self) -> bool:
        return self == Mode.CLIENT

    @property
    def is_register(self) -> bool:
        return self == Mode.REGISTER

    @property
    def is_login_test(self) -> bool:
        return self == Mode.LOGIN_TEST

    @property
    def is_anonymous_test(self) -> bool:
        return self == Mode.ANONYMOUS_TEST


class HTTPRequestError(IntEnum):
    UNKNOWN = 0
    INCOMPLETE = 1
    STATUS_NOT_OK = 2
    CANCELLED = 3
    CONTENT_OVERFLOW = 4
    TIMEOUT = 5

    def __str__(self) -> str:
        return self.name


MOODS = [
    "afraid",
    "amazed",
    "amorous",
    "angry",
    "annoyed",
    "anxious",
    "aroused",
    "ashamed",
    "bored",
    "brave",
    "calm",
    "cautious",
    "cold",
    "confident",
    "confused",
    "contemplative",
    "contented",
    "cranky",
    "crazy",
    "creative",
    "curious",
    "dejected",
    "depressed",
    "disappointed",
    "disgusted",
    "dismayed",
    "distracted",
    "embarrassed",
    "envious",
    "excited",
    "flirtatious",
    "frustrated",
    "grateful",
    "grieving",
    "grumpy",
    "guilty",
    "happy",
    "hopeful",
    "hot",
    "humbled",
    "humiliated",
    "hungry",
    "hurt",
    "impressed",
    "in_awe",
    "in_love",
    "indignant",
    "interested",
    "intoxicated",
    "invincible",
    "jealous",
    "lonely",
    "lost",
    "lucky",
    "mean",
    "moody",
    "nervous",
    "neutral",
    "offended",
    "outraged",
    "playful",
    "proud",
    "relaxed",
    "relieved",
    "remorseful",
    "restless",
    "sad",
    "sarcastic",
    "satisfied",
    "serious",
    "shocked",
    "shy",
    "sick",
    "sleepy",
    "spontaneous",
    "stressed",
    "strong",
    "surprised",
    "thankful",
    "thirsty",
    "tired",
    "undefined",
    "weak",
    "worried",
]


ACTIVITIES = {
    "doing_chores": [
        "buying_groceries",
        "cleaning",
        "cooking",
        "doing_maintenance",
        "doing_the_dishes",
        "doing_the_laundry",
        "gardening",
        "running_an_errand",
        "walking_the_dog",
    ],
    "drinking": ["having_a_beer", "having_coffee", "having_tea"],
    "eating": ["having_a_snack", "having_breakfast", "having_dinner", "having_lunch"],
    "exercising": [
        "cycling",
        "dancing",
        "hiking",
        "jogging",
        "playing_sports",
        "running",
        "skiing",
        "swimming",
        "working_out",
    ],
    "grooming": [
        "at_the_spa",
        "brushing_teeth",
        "getting_a_haircut",
        "shaving",
        "taking_a_bath",
        "taking_a_shower",
    ],
    "having_appointment": [],
    "inactive": [
        "day_off",
        "hanging_out",
        "hiding",
        "on_vacation",
        "praying",
        "scheduled_holiday",
        "sleeping",
        "thinking",
    ],
    "relaxing": [
        "fishing",
        "gaming",
        "going_out",
        "partying",
        "reading",
        "rehearsing",
        "shopping",
        "smoking",
        "socializing",
        "sunbathing",
        "watching_tv",
        "watching_a_movie",
    ],
    "talking": ["in_real_life", "on_the_phone", "on_video_phone"],
    "traveling": [
        "commuting",
        "cycling",
        "driving",
        "in_a_car",
        "on_a_bus",
        "on_a_plane",
        "on_a_train",
        "on_a_trip",
        "walking",
    ],
    "working": ["coding", "in_a_meeting", "studying", "writing"],
}


LOCATION_DATA = [
    "accuracy",
    "alt",
    "altaccuracy",
    "area",
    "bearing",
    "building",
    "country",
    "countrycode",
    "datum",
    "description",
    "error",
    "floor",
    "lat",
    "locality",
    "lon",
    "postalcode",
    "region",
    "room",
    "speed",
    "street",
    "text",
    "timestamp",
    "tzo",
    "uri",
]


TUNE_DATA = ["artist", "length", "rating", "source", "title", "track", "uri"]


REGISTER_FIELDS = [
    "username",
    "nick",
    "password",
    "name",
    "first",
    "last",
    "email",
    "address",
    "city",
    "state",
    "zip",
    "phone",
    "url",
    "date",
]

GIO_TLS_ERRORS = {
    Gio.TlsCertificateFlags.UNKNOWN_CA: "The signing certificate authority is not known",
    Gio.TlsCertificateFlags.REVOKED: "The certificate has been revoked",
    Gio.TlsCertificateFlags.BAD_IDENTITY: "The certificate does not match the expected identity of the site",
    Gio.TlsCertificateFlags.INSECURE: "The certificate’s algorithm is insecure",
    Gio.TlsCertificateFlags.NOT_ACTIVATED: "The certificate’s activation time is in the future",
    Gio.TlsCertificateFlags.GENERIC_ERROR: "Unknown validation error",
    Gio.TlsCertificateFlags.EXPIRED: "The certificate has expired",
}

NOT_ALLOWED_XML_CHARS = {
    "&": "&amp;",
    "<": "&lt;",
    ">": "&gt;",
    '"': "&quot;",
    "\x0C": "",
    "\x1B": "",
}


SOUP_ENCODING: set[Soup.Encoding] = {
    Soup.Encoding.UNRECOGNIZED,
    Soup.Encoding.NONE,
    Soup.Encoding.CONTENT_LENGTH,
    Soup.Encoding.EOF,
    Soup.Encoding.CHUNKED,
    Soup.Encoding.BYTERANGES,
}



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import ParamSpec
from typing import TYPE_CHECKING
from typing import TypeVar

import functools
import inspect
import logging
from collections.abc import Callable
from urllib.parse import unquote
from urllib.parse import urlparse

from nbxmpp.errors import is_error
from nbxmpp.errors import StanzaError
from nbxmpp.protocol import Iq
from nbxmpp.simplexml import Node
from nbxmpp.structs import CommonResult

if TYPE_CHECKING:
    from nbxmpp.dispatcher import NBXMPPModuleT
    from nbxmpp.task import Task

T = TypeVar("T")
P = ParamSpec("P")


def process_response(response: Iq) -> CommonResult:
    if response.isError():
        raise StanzaError(response)

    return CommonResult(jid=response.getFrom())


def raise_if_error(result: Any) -> None:
    if is_error(result):
        raise result


def finalize(task: Task, result: Any) -> Any:
    if is_error(result):
        raise result
    if isinstance(result, Node):
        return task.set_result(result)
    return result


def parse_xmpp_uri(uri: str) -> tuple[str, str, dict[str, str]]:
    url = urlparse(uri)
    if url.scheme != "xmpp":
        raise ValueError("not a xmpp uri")

    if ";" not in url.query:
        return (url.path, url.query, {})

    action, query = url.query.split(";", 1)
    key_value_pairs = query.split(";")

    dict_: dict[str, str] = {}
    for key_value in key_value_pairs:
        key, value = key_value.split("=")
        dict_[key] = unquote(value)

    return (url.path, action, dict_)


def make_func_arguments_string(
    func: Callable[..., Any], self: Any, args: Any, kwargs: Any
) -> str:
    signature = inspect.signature(func)
    bound_arguments = signature.bind(self, *args, **kwargs)
    bound_arguments.apply_defaults()
    arg_string = ""
    for name, arg in bound_arguments.arguments.items():
        if name == "self":
            continue
        arg_string += f"{name}={arg}, "
    arg_string = arg_string[:-2]
    return f"{func.__name__}({arg_string})"


def log_calls(func: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(func)
    def func_wrapper(self: NBXMPPModuleT, *args: Any, **kwargs: Any) -> T:
        if self._log.isEnabledFor(logging.INFO):  # type: ignore
            self._log.info(make_func_arguments_string(func, self, args, kwargs))  # type: ignore
        return func(self, *args, **kwargs)

    return func_wrapper



# simplexml.py based on Mattew Allum's xmlstream.py
#
# Copyright (C) 2003-2005 Alexey "Snake" Nezhdanov
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Simplexml module provides xmpppy library with all needed tools to handle XML
nodes and XML streams. I'm personally using it in many other separate
projects. It is designed to be as standalone as possible
"""

from __future__ import annotations

from typing import Any
from typing import Union

import logging
import xml.parsers.expat
from collections.abc import Callable
from collections.abc import Iterator
from copy import deepcopy
from xml.parsers.expat import ExpatError

from nbxmpp.const import NOT_ALLOWED_XML_CHARS

Attrs = dict[str, str]

log = logging.getLogger("nbxmpp.simplexml")


def XMLescape(text: str) -> str:
    """
    Return escaped text
    """

    for key, value in NOT_ALLOWED_XML_CHARS.items():
        text = text.replace(key, value)
    return text


class Node:
    """
    Node class describes syntax of separate XML Node. It have a constructor that
    permits node creation from set of "namespace name", attributes and payload
    of text strings and other nodes. It does not natively support building node
    from text string and uses NodeBuilder class for that purpose. After
    creation node can be mangled in many ways so it can be completely changed.
    Also node can be serialised into string in one of two modes: default (where
    the textual representation of node describes it exactly) and "fancy" - with
    whitespace added to make indentation and thus make result more readable by
    human.

    Node class have attribute FORCE_NODE_RECREATION that is defaults to False
    thus enabling fast node replication from the some other node. The drawback
    of the fast way is that new node shares some info with the "original" node
    that is changing the one node may influence the other. Though it is rarely
    needed (in xmpppy it is never needed at all since I'm usually never using
    original node after replication (and using replication only to move upwards
    on the classes tree).
    """

    name: str
    namespace: str
    attrs: Attrs
    data: list[str]
    kids: list[Node | str]
    parent: Node | None
    nsd: dict[str, str]
    nsp_cache: dict[Any, Any]

    FORCE_NODE_RECREATION = False

    def __init__(
        self,
        tag: str | None = None,
        attrs: Attrs | None = None,
        payload: Node | str | list[Node | str] | None = None,
        parent: Node | None = None,
        nsp: dict[Any, Any] | None = None,
        node_built: bool = False,
        node: Node | Any | None = None,
    ) -> None:
        """
        Takes "tag" argument as the name of node (prepended by namespace, if
        needed and separated from it by a space), attrs dictionary as the set of
        arguments, payload list as the set of textual strings and child nodes
        that this node carries within itself and "parent" argument that is
        another node that this one will be the child of. Also the __init__ can
        be provided with "node" argument that is either a text string containing
        exactly one node or another Node instance to begin with. If both "node"
        and other arguments is provided then the node initially created as
        replica of "node" provided and then modified to be compliant with other
        arguments.
        """
        if node:
            if self.FORCE_NODE_RECREATION and isinstance(node, Node):
                node = str(node)
            if not isinstance(node, Node):
                node = NodeBuilder(node, self)
                node_built = True
            else:
                self.name = node.name
                self.namespace = node.namespace
                self.attrs = {}
                self.data: list[str] = []
                self.kids = []
                self.parent = node.parent
                self.nsd = {}
                for key in node.attrs:
                    self.attrs[key] = node.attrs[key]
                for data in node.data:
                    self.data.append(data)
                for kid in node.kids:
                    self.kids.append(kid)
                for key, value in node.nsd.items():
                    self.nsd[key] = value
        else:
            self.name = "tag"
            self.namespace = ""
            self.attrs = {}
            self.data = []
            self.kids = []
            self.parent = None
            self.nsd = {}
        if parent:
            self.parent = parent
        self.nsp_cache = {}
        if nsp:
            for key, value in nsp.items():
                self.nsp_cache[key] = value

        if attrs is not None:
            for attr, val in attrs.items():
                if attr == "xmlns":
                    self.nsd[""] = val
                elif attr.startswith("xmlns:"):
                    self.nsd[attr[6:]] = val
                self.attrs[attr] = attrs[attr]

        if tag:
            if node_built:
                pfx, self.name = ([""] + tag.split(":"))[-2:]
                self.namespace = self.lookup_nsp(pfx)
            else:
                if " " in tag:
                    self.namespace, self.name = tag.split()
                else:
                    self.name = tag
        if payload is not None:
            if not isinstance(payload, list):
                payload = [payload]
            for i in payload:
                if isinstance(i, Node):
                    self.addChild(node=i)
                else:
                    self.data.append(str(i))

    def lookup_nsp(self, pfx: str = "") -> str:
        ns = self.nsd.get(pfx, None)
        if ns is None:
            ns = self.nsp_cache.get(pfx, None)
        if ns is None:
            if self.parent:
                ns = self.parent.lookup_nsp(pfx)
                self.nsp_cache[pfx] = ns
            else:
                return "http://www.gajim.org/xmlns/undeclared"
        return ns

    def __str__(self, fancy: int = 0) -> str:
        """
        Method used to dump node into textual representation. If "fancy"
        argument is set to True produces indented output for readability
        """
        s = (fancy - 1) * 2 * " " + "<" + self.name
        if self.namespace:
            if not self.parent or self.parent.namespace != self.namespace:
                if "xmlns" not in self.attrs:
                    s += ' xmlns="%s"' % self.namespace
        for key in self.attrs:
            val = str(self.attrs[key])
            s += ' %s="%s"' % (key, XMLescape(val))

        s += ">"
        cnt = 0
        if self.kids:
            if fancy:
                s += "\n"
            for a in self.kids:
                if not fancy and (len(self.data) - 1) >= cnt:
                    s += XMLescape(self.data[cnt])
                elif (len(self.data) - 1) >= cnt:
                    s += XMLescape(self.data[cnt].strip())
                if isinstance(a, str):
                    s += a.__str__()
                else:
                    s += a.__str__(fancy and fancy + 1)
                cnt += 1
        if not fancy and (len(self.data) - 1) >= cnt:
            s += XMLescape(self.data[cnt])
        elif (len(self.data) - 1) >= cnt:
            s += XMLescape(self.data[cnt].strip())
        if not self.kids and s.endswith(">"):
            s = s[:-1] + " />"
            if fancy:
                s += "\n"
        else:
            if fancy and not self.data:
                s += (fancy - 1) * 2 * " "
            s += "</" + self.name + ">"
            if fancy:
                s += "\n"
        return s

    def addChild(
        self,
        name: str | None = None,
        attrs: Attrs | None = None,
        payload: list[Any] | None = None,
        namespace: str | None = None,
        node: Node | None = None,
    ) -> Node:
        """
        If "node" argument is provided, adds it as child node. Else creates new
        node from the other arguments' values and adds it as well
        """
        if payload is None:
            payload = []

        if attrs is None:
            attrs = {}
        elif "xmlns" in attrs:
            raise AttributeError("Use namespace=x instead of attrs={'xmlns':x}")
        if node:
            newnode = node
            node.parent = self
        else:
            newnode = Node(tag=name, parent=self, attrs=attrs, payload=payload)
        if namespace:
            newnode.setNamespace(namespace)
        self.kids.append(newnode)
        return newnode

    def addData(self, data: Any) -> None:
        """
        Add some CDATA to node
        """
        self.data.append(str(data))

    def clearData(self) -> None:
        """
        Remove all CDATA from the node
        """
        self.data = []

    def delAttr(self, key: str) -> None:
        """
        Delete an attribute "key"
        """
        del self.attrs[key]

    def delChild(self, node: Node | str, attrs: Attrs | None = None) -> Node | None:
        """
        Delete the "node" from the node's childs list, if "node" is an instance.
        Else delete the first node that have specified name and (optionally)
        attributes
        """
        if not isinstance(node, Node):
            node = self.getTag(node, attrs)

        self.kids.remove(node)
        return node

    def getAttrs(self, copy: bool = False) -> Attrs:
        """
        Return all node's attributes as dictionary
        """
        if copy:
            return deepcopy(self.attrs)
        return self.attrs

    def getAttr(self, key: str) -> str | None:
        """
        Return value of specified attribute
        """
        return self.attrs.get(key)

    def getNamespacedAttr(self, key: str, namespace: str) -> str | None:
        """
        Return an attribute of this Node using a different namespace than the
        node's main namespace.

        NB: This is not a full implementation. It does not search for prefixes defined
        in parent elements, only for prefixes defined in this `Node`.
        """
        for prefix, ns in self.nsd.items():
            if ns != namespace:
                continue
            attr = self.attrs.get(f"{prefix}:{key}")
            if attr is None:
                # there might be several prefixes defined for a given namespace
                continue
            return attr
        return None

    def getChildren(self) -> list[Node | str]:
        """
        Return all node's child nodes as list
        """
        return self.kids

    def getData(self) -> str:
        """
        Return all node CDATA as string (concatenated)
        """
        return "".join(self.data)

    def getName(self) -> str:
        """
        Return the name of node
        """
        return self.name

    def getNamespace(self) -> str:
        """
        Return the namespace of node
        """
        return self.namespace

    def getParent(self) -> Node | None:
        """
        Returns the parent of node (if present)
        """
        return self.parent

    def getTag(
        self, name: str, attrs: Attrs | None = None, namespace: str | None = None
    ) -> Node | None:
        """
        Filter all child nodes using specified arguments as filter. Return the
        first found or None if not found
        """
        tags = self.getTags(name, attrs, namespace)
        if not tags:
            return None
        return tags[0]

    def getTagAttr(
        self, tag: str, attr: str, namespace: str | None = None
    ) -> str | None:
        """
        Return attribute value of the child with specified name (or None if no
        such attribute)
        """
        node = self.getTag(tag, namespace=namespace)
        if node is None:
            return None
        return node.getAttr(attr)

    def getTagData(self, tag: str) -> str | None:
        """
        Return concatenated CDATA of the child with specified name
        """
        node = self.getTag(tag)
        if node is None:
            return None
        return node.getData()

    def getTags(
        self, name: str, attrs: Attrs | None = None, namespace: str | None = None
    ) -> list[Node]:
        """
        Filter all child nodes using specified arguments as filter. Returns the
        list of nodes found
        """
        nodes: list[Node] = []
        for node in self.kids:
            if namespace and namespace != node.getNamespace():
                continue
            if node.getName() == name:
                if attrs is None:
                    attrs = {}
                for key in attrs:
                    if key not in node.attrs or node.attrs[key] != attrs[key]:
                        break
                else:
                    nodes.append(node)
        return nodes

    def iterTags(
        self, name: str, attrs: Attrs | None = None, namespace: str | None = None
    ) -> Iterator[Node]:
        """
        Iterate over all children using specified arguments as filter
        """
        for node in self.kids:
            if namespace is not None and namespace != node.getNamespace():
                continue
            if node.getName() == name:
                if attrs is None:
                    attrs = {}
                for key in attrs:
                    if key not in node.attrs or node.attrs[key] != attrs[key]:
                        break
                else:
                    yield node

    def setAttr(self, key: str, val: str) -> None:
        """
        Set attribute "key" with the value "val"
        """
        self.attrs[key] = val

    def setData(self, data: Any) -> None:
        """
        Set node's CDATA to provided string. Resets all previous CDATA!
        """
        self.data = [str(data)]

    def setName(self, val: str) -> None:
        """
        Change the node name
        """
        self.name = val

    def setNamespace(self, namespace: str) -> None:
        """
        Changes the node namespace
        """
        self.namespace = namespace

    def setParent(self, node: Node) -> None:
        """
        Set node's parent to "node". WARNING: do not checks if the parent
        already present and not removes the node from the list of childs of
        previous parent
        """
        self.parent = node

    def setPayload(
        self, payload: list[Node | str] | Node | str, add: bool = False
    ) -> None:
        """
        Set node payload according to the list specified. WARNING: completely
        replaces all node's previous content. If you wish just to add child or
        CDATA - use addData or addChild methods
        """
        if not isinstance(payload, list):
            payload = [payload]
        if add:
            self.kids += payload
        else:
            self.kids = payload

    def setTag(
        self, name: str, attrs: Attrs | None = None, namespace: str | None = None
    ) -> Node:
        """
        Same as getTag but if the node with specified namespace/attributes not
        found, creates such node and returns it
        """
        nodes = self.getTags(name, attrs, namespace=namespace)
        if nodes:
            return nodes[0]
        return self.addChild(name, attrs, namespace=namespace)

    def setTagAttr(
        self, tag: str, attr: str, val: str, namespace: str | None = None
    ) -> None:
        """
        Create new node (if not already present) with name "tag" and set it's
        attribute "attr" to value "val"
        """
        try:
            self.getTag(tag, namespace=namespace).attrs[attr] = val
        except Exception:
            self.addChild(tag, namespace=namespace, attrs={attr: val})

    def setTagData(self, tag: str, val: Any, attrs: Attrs | None = None) -> None:
        """
        Creates new node (if not already present) with name "tag" and
        (optionally) attributes "attrs" and sets it's CDATA to string "val"
        """
        try:
            self.getTag(tag, attrs).setData(str(val))
        except Exception:
            self.addChild(tag, attrs, payload=[str(val)])

    def getXmlLang(self) -> str | None:
        lang = self.attrs.get("xml:lang")
        if lang is not None:
            return lang

        if self.parent is not None:
            return self.parent.getXmlLang()
        return None

    def has_attr(self, key: str) -> bool:
        """
        Check if node have attribute "key"
        """
        return key in self.attrs

    def __getitem__(self, item: str) -> str | None:
        """
        Return node's attribute "item" value
        """
        return self.getAttr(item)

    def __setitem__(self, item: str, val: str) -> None:
        """
        Set node's attribute "item" value
        """
        self.setAttr(item, val)

    def __delitem__(self, item: str) -> None:
        """
        Delete node's attribute "item"
        """
        self.delAttr(item)

    def __contains__(self, item: str) -> bool:
        """
        Check if node has attribute "item"
        """
        return self.has_attr(item)

    def __getattr__(self, attr: str) -> Union["T" | "NT"]:
        """
        Reduce memory usage caused by T/NT classes - use memory only when needed
        """
        if attr == "T":
            self.T = T(self)
            return self.T
        if attr == "NT":
            self.NT = NT(self)
            return self.NT
        raise AttributeError


class T:
    """
    Auxiliary class used to quick access to node's child nodes
    """

    def __init__(self, node):
        self.__dict__["node"] = node

    def __getattr__(self, attr):
        return self.node.setTag(attr)

    def __setattr__(self, attr, val):
        if isinstance(val, Node):
            Node.__init__(self.node.setTag(attr), node=val)
            return None
        return self.node.setTagData(attr, val)

    def __delattr__(self, attr):
        return self.node.delChild(attr)


class NT(T):
    """
    Auxiliary class used to quick create node's child nodes
    """

    def __getattr__(self, attr):
        return self.node.addChild(attr)

    def __setattr__(self, attr, val):
        if isinstance(val, Node):
            self.node.addChild(attr, node=val)
            return None
        return self.node.addChild(attr, payload=[val])


class NodeBuilder:
    """
    Builds a Node class minidom from data parsed to it. This class used for two
    purposes:

      1. Creation an XML Node from a textual representation. F.e. reading a
         config file. See an XML2Node method.
      2. Handling an incoming XML stream. This is done by mangling the
         __dispatch_depth parameter and redefining the dispatch method.

    You do not need to use this class directly if you do not designing your own
    XML handler
    """

    _parser: Any
    Parse: Callable[[str, bool], None]
    __depth: int
    __max_depth: int
    _dispatch_depth: int
    _document_attrs: Attrs | None
    _document_nsp: dict[str, str] | None
    _mini_dom: Node | None
    last_is_data: bool
    _ptr: Node | None
    data_buffer: list[str] | None
    stream_error: str
    _is_stream: bool

    def __init__(
        self,
        data: str | None = None,
        initial_node: Node | None = None,
        dispatch_depth: int = 1,
        finished: bool = True,
    ) -> None:
        """
        Take two optional parameters: "data" and "initial_node"

        By default class initialised with empty Node class instance. Though, if
        "initial_node" is provided it used as "starting point". You can think
        about it as of "node upgrade". "data" (if provided) feeded to parser
        immidiatedly after instance init.
        """
        self._parser = xml.parsers.expat.ParserCreate()
        self._parser.UseForeignDTD(False)
        self._parser.StartElementHandler = self.starttag
        self._parser.EndElementHandler = self.endtag
        self._parser.StartNamespaceDeclHandler = self.handle_namespace_start
        self._parser.CharacterDataHandler = self.handle_cdata
        self._parser.StartDoctypeDeclHandler = self.handle_invalid_xmpp_element
        self._parser.EntityDeclHandler = self.handle_invalid_xmpp_element
        self._parser.CommentHandler = self.handle_invalid_xmpp_element
        self._parser.ExternalEntityRefHandler = self.handle_invalid_xmpp_element
        self._parser.AttlistDeclHandler = self.handle_invalid_xmpp_element
        self._parser.ProcessingInstructionHandler = self.handle_invalid_xmpp_element
        self._parser.buffer_text = True
        self.Parse = self._parser.Parse

        self.__depth = 0
        self.__max_depth = 0
        self._dispatch_depth = dispatch_depth
        self._document_attrs = None
        self._document_nsp = None
        self._mini_dom = initial_node
        self.last_is_data = True
        self._ptr = None
        self.data_buffer = None
        self.stream_error = ""
        self._is_stream = not finished
        if data:
            self._parser.Parse(data, finished)

    def check_data_buffer(self) -> None:
        if self.data_buffer:
            self._ptr.data.append("".join(self.data_buffer))
            del self.data_buffer[:]
            self.data_buffer = None

    def destroy(self) -> None:
        """
        Method used to allow class instance to be garbage-collected
        """
        self.check_data_buffer()
        self._parser.StartElementHandler = None
        self._parser.EndElementHandler = None
        self._parser.CharacterDataHandler = None
        self._parser.StartNamespaceDeclHandler = None

    def starttag(self, tag: str, attrs: Attrs) -> None:
        """
        XML Parser callback. Used internally
        """
        self.check_data_buffer()
        self._inc_depth()
        log.debug(
            "STARTTAG.. DEPTH -> %i , tag -> %s, attrs -> %s", self.__depth, tag, attrs
        )
        if self.__depth == self._dispatch_depth:
            if not self._mini_dom:
                self._mini_dom = Node(
                    tag=tag, attrs=attrs, nsp=self._document_nsp, node_built=True
                )
            else:
                Node.__init__(
                    self._mini_dom,
                    tag=tag,
                    attrs=attrs,
                    nsp=self._document_nsp,
                    node_built=True,
                )
            self._ptr = self._mini_dom
        elif self.__depth > self._dispatch_depth:
            self._ptr.kids.append(
                Node(tag=tag, parent=self._ptr, attrs=attrs, node_built=True)
            )
            self._ptr = self._ptr.kids[-1]
        if self.__depth == 1:
            self._document_attrs = {}
            self._document_nsp = {}
            nsp, name = ([""] + tag.split(":"))[-2:]
            for attr, val in attrs.items():
                if attr == "xmlns":
                    self._document_nsp[""] = val
                elif attr.startswith("xmlns:"):
                    self._document_nsp[attr[6:]] = val
                else:
                    self._document_attrs[attr] = val
            ns = self._document_nsp.get(
                nsp, "http://www.gajim.org/xmlns/undeclared-root"
            )
            try:
                header = Node(
                    tag=tag, attrs=attrs, nsp=self._document_nsp, node_built=True
                )
                self.dispatch(header)
                self._check_stream_start(ns, name)
            except ValueError as error:
                self._document_attrs = None
                raise ValueError(str(error))
        if not self.last_is_data and self._ptr.parent:
            self._ptr.parent.data.append("")
        self.last_is_data = False

    def _check_stream_start(self, ns: str, tag: str) -> None:
        if self._is_stream:
            if ns != "http://etherx.jabber.org/streams" or tag != "stream":
                raise ValueError(
                    "Incorrect stream start: (%s,%s). Terminating." % (tag, ns)
                )
        else:
            self.stream_header_received()

    def endtag(self, tag: str) -> None:
        """
        XML Parser callback. Used internally
        """
        log.debug("DEPTH -> %i , tag -> %s", self.__depth, tag)
        self.check_data_buffer()
        if self.__depth == self._dispatch_depth:
            if self._mini_dom.getName() == "error":
                children = self._mini_dom.getChildren()
                if children:
                    self.stream_error = children[0].getName()
                else:
                    self.stream_error = self._mini_dom.getData()
            self.dispatch(self._mini_dom)
        elif self.__depth > self._dispatch_depth:
            self._ptr = self._ptr.parent
        else:
            log.debug("Got higher than dispatch level. Stream terminated?")
        self._dec_depth()
        self.last_is_data = False
        if self.__depth == 0:
            self.stream_footer_received()

    def handle_cdata(self, data: str) -> None:
        if self.last_is_data:
            if self.data_buffer:
                self.data_buffer.append(data)
        elif self._ptr:
            self.data_buffer = [data]
            self.last_is_data = True

    @staticmethod
    def handle_invalid_xmpp_element(*args: Any) -> None:
        raise ExpatError("Found invalid xmpp stream element: %s" % str(args))

    def handle_namespace_start(self, _prefix: str, _uri: str) -> None:
        """
        XML Parser callback. Used internally
        """
        self.check_data_buffer()

    def getDom(self) -> Node | None:
        """
        Return just built Node
        """
        self.check_data_buffer()
        return self._mini_dom

    def dispatch(self, stanza: Any) -> None:
        """
        Get called when the NodeBuilder reaches some level of depth on it's way
        up with the built node as argument. Can be redefined to convert incoming
        XML stanzas to program events
        """

    def stream_header_received(self) -> None:
        """
        Method called when stream just opened
        """
        self.check_data_buffer()

    def stream_footer_received(self) -> None:
        """
        Method called when stream just closed
        """
        self.check_data_buffer()

    def has_received_endtag(self, level: int = 0) -> bool:
        """
        Return True if at least one end tag was seen (at level)
        """
        return self.__depth <= level < self.__max_depth

    def _inc_depth(self) -> None:
        self.__depth += 1
        self.__max_depth = max(self.__depth, self.__max_depth)

    def _dec_depth(self) -> None:
        self.__depth -= 1


def XML2Node(xml_str: str) -> Node | None:
    """
    Convert supplied textual string into XML node. Handy f.e. for reading
    configuration file. Raises xml.parser.expat.parsererror if provided string
    is not well-formed XML
    """
    return NodeBuilder(xml_str).getDom()


def BadXML2Node(xml_str: str) -> Node | None:
    """
    Convert supplied textual string into XML node. Survives if xml data is
    cutted half way round. I.e. "<html>some text <br>some more text". Will raise
    xml.parser.expat.parsererror on misplaced tags though. F.e. "<b>some text
    <br>some more text</b>" will not work
    """
    return NodeBuilder(xml_str).getDom()
